# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Linear matrix inequalities."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class LMIConstraint(Constraint):
    """Linear matrix inequality.

    An inequality with respect to the positive semidefinite cone, also known as
    a Linear Matrix Inequality (LMI) or an SDP constraint.
    """

    def __init__(self, lhs, relation, rhs, customString=None):
        """Construct a :class:`LMIConstraint`.

        :param ~picos.expressions.AffineExpression lhs:
            Left hand side expression.
        :param str relation:
            Constraint relation symbol.
        :param ~picos.expressions.AffineExpression rhs:
            Right hand side expression.
        :param str customString:
            Optional string description.
        """
        from ..expressions import (AffineExpression, HermitianVariable,
                                   SymmetricVariable)
        from ..expressions.vectorizations import (HermitianVectorization,
                                                  SymmetricVectorization)
        from ..expressions.data import cvxopt_equals

        required_type = self._required_type()

        assert isinstance(lhs, required_type)
        assert isinstance(rhs, required_type)
        assert relation in self.LE + self.GE

        if lhs.shape != rhs.shape:
            raise ValueError("Failed to form a constraint: "
                "Expressions have incompatible dimensions.")

        if lhs.shape[0] != lhs.shape[1]:
            raise ValueError("Failed to form a constraint: "
                "LMI expressions are not square.")

        self.lhs      = lhs
        self.rhs      = rhs
        self.relation = relation

        psd = self.psd
        if not psd.hermitian:
            needed = "symmetric" if required_type is AffineExpression \
                else "hermitian"

            raise ValueError("Failed to form a constraint: {} is not "
                "necessarily {}. Consider a constraint on {} instead."
                .format(psd, needed, glyphs.Tr("{}.hermitianized")(psd.string)))

        # Check if the constraint simply poses positive semidefiniteness on a
        # matrix variable, as certain solvers can handle this more efficiently
        # than a general linear matrix inequality.
        self.semidefVar = None
        if len(psd._coefs) == 1 and not psd._const:
            var, coef = list(psd._coefs.items())[0]

            # Don't accept any numeric deviation as we don't know the user's
            # tolerance settings.
            if isinstance(var, SymmetricVariable) and cvxopt_equals(
                    coef, SymmetricVectorization(psd.shape).identity,
                    tolerance=0.0):
                self.semidefVar = var
            elif isinstance(var, HermitianVariable) and cvxopt_equals(
                    coef, HermitianVectorization(psd.shape).identity,
                    tolerance=0.0):
                self.semidefVar = var

        super(LMIConstraint, self).__init__(
            self._get_type_term(), customString, printSize=True)

    def _get_type_term(self):
        return "LMI"

    def _required_type(self):
        from ..expressions import AffineExpression

        return AffineExpression

    @property
    def smaller(self):
        """The smaller-or-equal side expression."""
        return self.rhs if self.relation == self.GE else self.lhs

    @property
    def greater(self):
        """The greater-or-equal side expression."""
        return self.lhs if self.relation == self.GE else self.rhs

    @cached_property
    def psd(self):
        """The matrix expression posed to be positive semidefinite."""
        if self.relation == self.GE:
            return self.lhs - self.rhs
        else:
            return self.rhs - self.lhs

    @cached_property
    def nsd(self):
        """The matrix expression posed to be negative semidefinite."""
        if self.relation == self.GE:
            return self.rhs - self.lhs
        else:
            return self.lhs - self.rhs

    nnd = psd
    npd = nsd

    Subtype = namedtuple("Subtype", ("diag",))

    def _subtype(self):
        return self.Subtype(self.lhs.shape[0])

    @classmethod
    def _cost(cls, subtype):
        n = subtype.diag
        return n*(n + 1)//2

    def _expression_names(self):
        yield "lhs"
        yield "rhs"

    def _str(self):
        if self.relation == self.LE:
            return glyphs.psdle(self.lhs.string, self.rhs.string)
        else:
            return glyphs.psdge(self.lhs.string, self.rhs.string)

    def _get_size(self):
        return self.lhs.shape

    def _get_slack(self):
        return self.psd.value


class ComplexLMIConstraint(LMIConstraint):
    """Complex linear matrix inequality."""

    class RealConversion(ConstraintConversion):
        """Complex LMI to real LMI conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            n = subtype.diag

            yield ("con", LMIConstraint.make_type(diag=2*n), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..modeling import Problem

            P = Problem()
            Z = con.psd
            P.add_constraint((Z.real & -Z.imag) // (Z.imag & Z.real) >> 0)

            return P

        @classmethod
        def dual(cls, auxVarPrimals, auxConDuals, options):
            """Implement :meth:`~.constraint.ConstraintConversion.dual`."""
            assert len(auxConDuals) == 1

            auxConDual = auxConDuals[0]
            if auxConDual is None:
                return None
            else:
                assert auxConDual.size[0] == auxConDual.size[1]
                n = auxConDual.size[0] // 2
                assert 2*n == auxConDual.size[0]
                A = auxConDual[:n, :n]
                B = auxConDual[:n, n:]
                D = auxConDual[n:, n:]
                return (A + 1j*B) + (D + 1j*B).H

    def _get_type_term(self):
        return "Complex LMI"

    def _required_type(self):
        from ..expressions import ComplexAffineExpression

        return ComplexAffineExpression

    @classmethod
    def _cost(cls, subtype):
        return subtype.diag**2


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`SumExponentials`."""

import math
import operator
from collections import namedtuple

import cvxopt
import numpy

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property, cached_unary_operator
from ..constraints import LogSumExpConstraint, SumExponentialsConstraint
from .data import convert_and_refine_arguments, convert_operands, cvx2np
from .exp_affine import AffineExpression
from .expression import Expression, refine_operands, validate_prediction

_API_START = api_start(globals())
# -------------------------------


class SumExponentials(Expression):
    r"""Sum of elementwise exponentials of an affine expression.

    :Definition:

    Let :math:`x` be an :math:`n`-dimensional real affine expression.

    1.  If no additional expression :math:`y` is given, this is the sum of
        elementwise exponentials

        .. math::

            \sum_{i = 1}^n \exp(\operatorname{vec}(x)_i).

    2.  If an additional affine expression :math:`y` of same shape as :math:`x`
        is given, this is the sum of elementwise perspectives of exponentials

        .. math::

            \sum_{i = 1}^n \operatorname{vec}(y)_i \exp\left(
            \frac{\operatorname{vec}(x)_i}{\operatorname{vec}(y)_i}\right).

    .. warning::

        When you pose an upper bound :math:`t` on a sum of elementwise
        exponentials, then PICOS enforces :math:`t \geq 0` through an auxiliary
        constraint during solution search. When an additional expression
        :math:`y` is given, PICOS enforces :math:`y \geq 0` as well.
    """

    # --------------------------------------------------------------------------
    # Initialization and factory methods.
    # --------------------------------------------------------------------------

    @convert_and_refine_arguments("x", "y", allowNone=True)
    def __init__(self, x, y=None):
        """Construct a :class:`SumExponentials`.

        :param x: The affine expression :math:`x`.
        :type x: ~picos.expressions.AffineExpression
        :param y: An additional affine expression :math:`y`. If necessary, PICOS
            will attempt to reshape or broadcast it to the shape of :math:`x`.
        :type y: ~picos.expressions.AffineExpression
        """
        if not isinstance(x, AffineExpression):
            raise TypeError("Can only sum the elementwise exponentials of a "
                "real affine expression, not of {}.".format(x.string))

        if y is not None:
            if not isinstance(y, AffineExpression):
                raise TypeError("The additional parameter y must be a real "
                    "affine expression, not {}.".format(y.string))
            elif x.shape != y.shape:
                y = y.reshaped_or_broadcasted(x.shape)

            if y.is1:
                y = None

        self._x = x
        self._y = y

        if len(x) == 1:
            if y is None:
                typeStr = "Exponential"
                symbStr = glyphs.exp(x.string)
            else:
                typeStr = "Exponential Perspective"
                symbStr = glyphs.mul(
                    y.string, glyphs.exp(glyphs.div(x.string, y.string)))
        else:
            if y is None:
                typeStr = "Sum of Exponentials"
                symbStr = glyphs.make_function("sum", "exp")(x.string)
            else:
                typeStr = "Sum of Exponential Perspectives"
                symbStr = glyphs.sum(glyphs.mul(glyphs.slice(y.string, "i"),
                    glyphs.exp(glyphs.div(glyphs.slice(x.string, "i"),
                    glyphs.slice(y.string, "i")))))

        Expression.__init__(self, typeStr, symbStr)

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    def _get_refined(self):
        if self._x.constant and (self._y is None or self._y.constant):
            return AffineExpression.from_constant(self.value, 1, self._symbStr)
        else:
            return self

    Subtype = namedtuple("Subtype", ("argdim", "y"))

    def _get_subtype(self):
        return self.Subtype(len(self._x), self._y is not None)

    def _get_value(self):
        x = numpy.ravel(cvx2np(self._x._get_value()))

        if self._y is None:
            s = numpy.sum(numpy.exp(x))
        else:
            y = numpy.ravel(cvx2np(self._y._get_value()))
            s = y.dot(numpy.exp(x / y))

        return cvxopt.matrix(s)

    @cached_unary_operator
    def _get_variables(self):
        if self._y is None:
            return self._x.variables
        else:
            return self._x.variables.union(self._y.variables)

    def _is_convex(self):
        return True

    def _is_concave(self):
        return False

    def _replace_variables(self, var_map):
        return self.__class__(self._x._replace_variables(var_map),
            None if self._y is None else self._y._replace_variables(var_map))

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __add__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only add a constant term to "
                    "a nonconstant PICOS sum of exponentials.")

            value = other.value

            if value < 0:
                raise NotImplementedError("You may only add a nonnegative term "
                    "to a nonconstant PICOS sum of exponentials.")

            if value == 0:
                # NOTE: We could return self here, but this is more consistent
                #       with other expressions' __add__ methods.
                sumexp = SumExponentials(self._x)
            elif self._y is None:
                sumexp = SumExponentials(self._x // math.log(value))
            else:
                sumexp = SumExponentials(self._x // value, self._y // 1)

            sumexp._typeStr = "Offset " + sumexp._typeStr
            sumexp._symbStr = glyphs.clever_add(self.string, other.string)

            return sumexp
        elif isinstance(other, SumExponentials):
            if self._y is None and other._y is None:
                sumexp = SumExponentials(self._x.vec // other._x.vec)
            elif self._y is not None and other._y is None:
                one = AffineExpression.from_constant(1.0, (other.n, 1))
                sumexp = SumExponentials(
                    self._x.vec // other._x.vec, self._y.vec // one)
            elif self._y is None and other._y is not None:
                one = AffineExpression.from_constant(1.0, (self.n, 1))
                sumexp = SumExponentials(
                    self._x.vec // other._x.vec, one // other._y.vec)
            else:
                sumexp = SumExponentials(
                    self._x.vec // other._x.vec, self._y.vec // other._y.vec)

            sumexp._symbStr = glyphs.clever_add(self.string, other.string)

            return sumexp
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __radd__(self, other):
        if isinstance(other, (AffineExpression, SumExponentials)):
            sumexp = self.__add__(other)
            # NOTE: __add__ always creates a fresh expression.
            sumexp._symbStr = glyphs.clever_add(other.string, self.string)
            return sumexp
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __mul__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS sum of exponentials with a constant term.")

            value = other.value

            if value < 0:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS sum of exponential with a nonnegative term.")

            if value == 0:
                return AffineExpression.zero()

            if self._y is None:
                sumexp = SumExponentials(self._x + math.log(value))
            else:
                sumexp = SumExponentials(self._x * value, self._y * value)

            sumexp._typeStr = "Scaled " + sumexp._typeStr
            sumexp._symbStr = glyphs.clever_mul(self.string, other.string)

            return sumexp
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __rmul__(self, other):
        if isinstance(other, AffineExpression):
            sumexp = self.__mul__(other)
            # NOTE: __mul__ always creates a fresh expression.
            sumexp._symbStr = glyphs.clever_mul(other.string, self.string)
            return sumexp
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __div__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only divide a nonconstant "
                    "PICOS sum of exponentials by a constant term.")

            value = other.value

            if value <= 0:
                raise NotImplementedError("You may only divide a nonconstant "
                    "PICOS sum of exponential by a positive term.")

            sumexp = self * (1.0 / value)
            # NOTE: __mul__ always creates a fresh expression.
            sumexp._symbStr = glyphs.div(self.string, other.string)
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Methods and properties that return expressions.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """The expression :math:`x`."""
        return self._x

    @property
    def y(self):
        """The additional expression :math:`y`, or :obj:`None`."""
        return self._y

    @cached_property
    def log(self):
        """The logarithm of the expression."""
        from . import LogSumExp

        if self._y is not None:
            raise NotImplementedError("May only take the logarithm of a sum of"
                " exponentials, not of a sum of exponential perspectives.")

        return LogSumExp(self._x)

    # --------------------------------------------------------------------------
    # Methods and properties that describe the expression.
    # --------------------------------------------------------------------------

    @property
    def n(self):
        """Length of :attr:`x`."""
        return len(self._x)

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        if relation == operator.__le__:
            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return SumExponentialsConstraint.make_type(
                    argdim=subtype.argdim,
                    lse_representable=(not subtype.y and other.subtype.nonneg))
            elif issubclass(other.clstype, SumExponentials):
                if subtype.y or other.subtype.y:
                    return NotImplemented

                if other.subtype.argdim != 1:
                    return NotImplemented

                return LogSumExpConstraint.make_type(argdim=subtype.argdim)

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __le__(self, other):
        from . import LogSumExp

        if isinstance(other, AffineExpression):
            return SumExponentialsConstraint(self, other)
        elif isinstance(other, SumExponentials):
            if self._y is not None or other._y is not None:
                raise NotImplementedError("Comparing two sums of exponentials "
                    "is not supported if either expression has the additional "
                    "perspectives parameter y set.")

            if other.n != 1:
                raise NotImplementedError("You may only upper bound a sum of "
                    "exponentials by a single exponential, not by another sum.")

            return LogSumExp(self._x) <= other._x
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""String templates used to print (algebraic) expressions.

PICOS internally uses this module to produce string representations for the
algebraic expressions that you create.
The function-like objects that are used to build such strings are called
"glyphs" and are instanciated by this module following the
`singleton pattern <https://en.wikipedia.org/wiki/Singleton_pattern>`_.
As a result, you can modify the glyph objects listed below to influence how
PICOS will format future strings, for example to disable use of unicode symbols
that your console does not suppport or to adapt PICOS' output to the rest of
your application.

Here's an example of first swapping the entire character set to display
expressions using only `Latin-1 <https://en.wikipedia.org/wiki/ISO/IEC_8859-1>`_
characters, and then modifying a single glyph to our liking:

  >>> import picos
  >>> X = picos.Problem().add_variable("X", (2,2), "symmetric")
  >>> print(X >> 0)
  X ≽ 0
  >>> picos.glyphs.latin1()
  >>> print(X >> 0)
  X » 0
  >>> picos.glyphs.psdge.template = "{} - {} is psd"
  >>> print(X >> 0)
  X - 0 is psd

Note that glyphs understand some algebraic rules such as operator precedence
and associativity. This is possible because strings produced by glyphs remember
how they were created.

  >>> one_plus_two = picos.glyphs.add(1, 2)
  >>> one_plus_two
  '1 + 2'
  >>> one_plus_two.glyph.template, one_plus_two.operands
  ('{} + {}', (1, 2))
  >>> picos.glyphs.add(one_plus_two, 3)
  '1 + 2 + 3'
  >>> picos.glyphs.sub(0, one_plus_two)
  '0 - (1 + 2)'

The positive semidefinite glyph above does not yet know how to properly handle
arguments with respect to the ``-`` symbol involved, but we can modify it
further:

  >>> print(X + X >> X + X)
  X + X - X + X is psd
  >>> # Use the same operator binding strength as regular substraction.
  >>> picos.glyphs.psdge.order = picos.glyphs.sub.order
  >>> print(X + X >> X + X)
  X + X - (X + X) is psd

You can reset all glyphs to their initial state as follows:

  >>> picos.glyphs.default()
"""

import functools
import sys

from .apidoc import api_end, api_start

# Allow functions to modify this module directly.
glyphs = sys.modules[__name__]

_API_START = api_start(globals())
# -------------------------------


# Definitions of glyph classes and the rich strings they create.
# --------------------------------------------------------------

class GlStr(str):
    """A string created from a :class:`glyph <Gl>`.

    It has an additional :attr:`glyph` field pointing to the glyph that created
    it, and a :attr:`operands` field containing the values used to create it.
    """

    def __new__(cls, string, glyph, operands):
        """Create a regular Python string."""
        return str.__new__(cls, string)

    def __init__(self, string, glyph, operands):
        """Augment the Python string with metadata on its origin."""
        self.glyph = glyph
        """The glyph used to create the string."""

        self.operands = operands
        """The operands used to create the string."""

    def __copy__(self):
        return self.__class__(str(self), self.glyph, self.operands)

    def reglyphed(self, replace={}):
        """Returns a rebuilt version of the string using current glyphs.

        :param dict replace:
            Replace leaf-node (non :class:`GlStr`) strings with new strings.
            This can be used, for instance, to change the names of varaibles.
        """
        return self.glyph(*(op.reglyphed(replace) if isinstance(op, GlStr) else
            (replace[op] if op in replace else op) for op in self.operands))


class Gl:
    """The basic "glyph", an (algebraic) string formatting template.

    Sublcasses are supposed to extend formatting routines, going beyond of what
    Python string formatting is capabale of. In particular, glyphs can be used
    to craft unambiguous algebraic expressions with the minimum amount of
    parenthesis.
    """

    def __init__(self, glyph):
        """Construct a glyph.

        :param str glyph: The glyph's format string template.
        """
        self.template = glyph
        self.initial  = glyph

    def reset(self):
        """Reset the glyph to its initial formatting template."""
        self.template = self.initial

    def update(self, new):
        """Change the glyph's formatting template."""
        self.template = new.template

    def rebuild(self):
        """If the template was created using other glyphs, rebuild it.

        :returns: True if the template has changed.
        """
        if isinstance(self.template, GlStr):
            oldTemplate = self.template
            self.template = self.template.reglyphed()
            return self.template != oldTemplate
        else:
            return False

    def __call__(self, *args):
        """Format the arguments as a :class:`GlStr`."""
        return GlStr(self.template.format(*args), self, args)


class OpStr(GlStr):
    """A string created from a math operator glyph."""

    pass


class Op(Gl):
    """The basic math operator glyph."""

    def __init__(self, glyph, order, assoc=False, closed=False):
        """Construct a math operator glyph.

        :param str glyph: The glyph's format string template.
        :param int order: The operator's position in the binding strength
            hierarchy. Operators with lower numbersbind more strongly.
        :param bool assoc: If this is :obj:`True`, then the operator is
            associative, so that parenthesis are always omitted around operands
            with an equal outer operator. Otherwise, (1) parenthesis are used
            around the right hand side operand of a binary operation of same
            binding strength and (2) around all operands of non-binary
            operations of same binding strength.
        :param closed: If :obj:`True`, the operator already encloses the
            operands in some sort of brackets, so that no additional parenthesis
            are needed. For glyphs where only some operands are enclosed, this
            can be specified per operand in the form of a list.
        :type closed: bool or list(bool)
        """
        self.initial = (glyph, order, assoc, closed)
        self.reset()

    def reset(self):
        """Reset the glyph to its initial behavior."""
        self.template, self.order, self.assoc, self.closed = self.initial

    def update(self, new):
        """Change the glyph's behavior."""
        self.template = new.template
        self.order    = new.order
        self.assoc    = new.assoc
        self.closed   = new.closed

    def __call__(self, *operands):
        """Format the arguments as an :class:`OpStr`."""
        if self.closed is True:
            return OpStr(self.template.format(*operands), self, operands)

        placeholders = []
        for i, operand in enumerate(operands):
            if isinstance(self.closed, list) and i < len(self.closed) \
            and self.closed[i]:
                parenthesis = False
            elif not isinstance(operand, OpStr):
                parenthesis = False
            elif operand.glyph.order < self.order:
                parenthesis = False
            elif operand.glyph.order == self.order:
                if len(operands) == 2 and i == 0:
                    # By default, bind from left to right.
                    parenthesis = False
                elif self.assoc in (None, False):
                    parenthesis = True
                else:
                    parenthesis = operand.glyph is not self
            else:
                parenthesis = True

            if type(operand) is float:
                # If no format specifier was given, then integral floats would
                # be formatted with a trailing '.0', which we don't want. Note
                # that for complex numbers the default behavior is already as we
                # want it, while 'g' would omit the parenthesis that we need.
                placeholder = "{:g}"
            else:
                placeholder = "{}"

            if parenthesis:
                placeholders.append(glyphs.parenth(placeholder))
            else:
                placeholders.append(placeholder)

        return OpStr(self.template.format(*placeholders).format(*operands),
            self, operands)


class Am(Op):
    """A math atom glyph."""

    def __init__(self, glyph):
        """Construct an :class:`Am` glyph.

        :param str glyph: The glyph's format string template.
        """
        Op.__init__(self, glyph, 0)


class Br(Op):
    """A math operator glyph with enclosing brackets."""

    def __init__(self, glyph):
        """Construct a :class:`Br` glyph.

        :param str glyph: The glyph's format string template.
        """
        Op.__init__(self, glyph, 0, closed=True)


class Fn(Op):
    """A math operator glyph in function form."""

    def __init__(self, glyph):
        """Construct a :class:`Fn` glyph.

        :param str glyph: The glyph's format string template.
        """
        Op.__init__(self, glyph, 0, closed=True)


class Tr(Op):
    """A math glyph in superscript/trailer form."""

    def __init__(self, glyph):
        """Construct a :class:`Tr` glyph.

        :param str glyph: The glyph's format string template.
        """
        Op.__init__(self, glyph, 1)


class Rl(Op):
    """A math relation glyph."""

    def __init__(self, glyph):
        """Construct a :class:`Rl` glyph.

        :param str glyph: The glyph's format string template.
        """
        Op.__init__(self, glyph, 5, assoc=True)


# Functions that show, reset or rebuild the glyph objects.
# --------------------------------------------------------

def show(*args):
    """Show output from all glyphs.

    :param list(str) args: Strings to use as glyph operands.
    """
    args = list(args) + ["{}"]*4

    print("{:8} | {:3} | {:5} | {}\n{}+{}+{}+{}".format(
        "Glyph", "Pri", "Asso", "Value", "-"*9, "-"*5, "-"*7, "-"*10))

    for name in sorted(list(glyphs.__dict__.keys())):
        glyph = getattr(glyphs, name)
        if isinstance(glyph, Gl):
            order = glyph.order if hasattr(glyph, "order") else ""
            assoc = str(glyph.assoc) if hasattr(glyph, "order") else ""
            print("{:8} | {:3} | {:5} | {}".format(
                name, order, assoc, glyph(*args)))


def rebuild():
    """Update glyphs that are based upon other glyphs."""
    for i in range(100):
        if not any(glyph.rebuild() for glyph in glyphs.__dict__.values()
                if isinstance(glyph, Gl)):
            return

    raise Exception("Maximum recursion depth for glyph rebuilding reached. "
        "There is likely a cyclic dependence between them.")


def ascii():
    """Let PICOS create future strings using only ASCII characters."""
    for glyph in glyphs.__dict__.values():
        if isinstance(glyph, Gl):
            glyph.reset()


# Initial glyph definitions and functions that update them.
# ---------------------------------------------------------

# Non-operator glyphs.
repr1    = Gl("<{}>"); """Representation glyph."""
repr2    = Gl(glyphs.repr1("{}: {}")); """Long representation glyph."""
parenth  = Gl("({})"); """Parenthesis glyph."""
sep      = Gl("{} : {}"); """Seperator glyph."""
compsep  = Gl("{}:{}"); """Compact seperator glyph."""
comma    = Gl("{}, {}"); """Seperated by comma glyph."""
size     = Gl("{}x{}"); """Matrix size/shape glyph."""
compose  = Gl("{}.{}"); """Function composition glyph."""
set      = Gl("{{{}}}"); """Set glyph."""
closure  = Fn("cl{}"); """Set closure glyph."""
interval = Gl("[{}]"); """Interval glyph."""
fromto   = Gl("{}..{}"); """Range glyph."""
intrange = Gl(glyphs.interval(glyphs.fromto("{}", "{}")))
"""Integer range glyph."""
shortint = Gl(glyphs.interval(glyphs.fromto("{},", ",{}")))
"""Shortened interval glyph."""
forall   = Gl("{} f.a. {}"); """Universal quantification glyph."""
leadsto  = Gl("{} -> {}"); """Successorship glyph."""
and_     = Gl("{} and {}"); """Logical and glyph."""
or_      = Gl("{} or {}"); """Logical or glyph."""

# Atomic glyphs.
infty    = Am("inf"); """Infinity glyph."""
idmatrix = Am("I"); """Identity matrix glyph."""
lambda_  = Am("lambda"); """Lambda symbol glyph."""

# Bracketed glyphs.
matrix   = Br("[{}]"); """Matrix glyph."""
dotp     = Br("<{}, {}>"); """Scalar product glyph."""
abs      = Br("|{}|"); """Absolute value glyph."""
norm     = Br("||{}||"); """Norm glyph."""

# Special norms.
pnorm    = Op(Gl("{}_{}")(glyphs.norm("{}"), "{}"), 1, closed=[True, False])
"""p-Norm glyph."""
pqnorm   = Op(Gl("{}_{},{}")(glyphs.norm("{}"), "{}", "{}"), 1,
    closed=[True, False, False]); """pq-Norm glyph."""
spnorm   = Op(Gl("{}_{}")(glyphs.norm("{}"), "2"), 1, closed=True)
"""Spectral Norm glyph."""
ncnorm   = Op(Gl("{}_{}")(glyphs.norm("{}"), "*"), 1, closed=True)
"""Nuclear Norm glyph."""

# Function glyphs.
sum      = Fn("sum({})");   """Summation glyph."""
max      = Fn("max({})");   """Maximum glyph."""
min      = Fn("min({})");   """Minimum glyph."""
exp      = Fn("exp({})");   """Exponentiation glyph."""
log      = Fn("log({})");   """Logarithm glyph."""
vec      = Fn("vec({})");   """Vectorization glyph."""
trilvec  = Fn("trilvec({})");  """Lower triangular vectorization glyph."""
triuvec  = Fn("triuvec({})");  """Upper triangular vectorization glyph."""
trace    = Fn("tr({})");    """Matrix trace glyph."""
diag     = Fn("diag({})");  """Diagonal matrix glyph."""
maindiag = Fn("maindiag({})");  """Main diagonal glyph."""
det      = Fn("det({})");   """Determinant glyph."""
real     = Fn("Re({})");    """Real part glyph."""
imag     = Fn("Im({})");    """Imaginary part glyph."""
conj     = Fn("conj({})");  """Complex conugate glyph."""
quadpart = Fn("quad({})");  """Quadratic part glyph."""
affpart  = Fn("aff({})");   """Affine part glyph."""
linpart  = Fn("lin({})");   """Linear part glyph."""
cstpart  = Fn("cst({})");   """Constant part glyph."""
value    = Fn("value({})"); """Value of expression glyph."""
reshaped = Fn("reshaped({}, {})"); """Reshaped glyph."""
bcasted  = Fn("bcasted({}, {})");  """Broadcasted glyph."""
sqrt     = Fn("sqrt({})");  """Square root glyph."""
shuffled = Fn("shuffled({})");     """Matrix reshuffling glyph."""

# Semi-closed glyphs.
ptrace   = Op("trace_{}({})", 0, closed=[False, True])
"""Matrix p-Trace glyph."""
slice    = Op("{}[{}]", 0, closed=[False, True])
"""Expression slicing glyph."""
ptransp_ = Op("{}.{{{{{}}}}}", 1, closed=[False, True])
"""Matrix partial transposition glyph."""  # TODO: Replace ptransp.
ptrace_  = Op("{}.{{{{{}}}}}", 1, closed=[False, True])
"""Matrix partial trace glyph."""  # TODO: Replace ptrace_.

# Trailer glyphs.
power    = Tr("{}^{}"); """Power glyph."""
cubed    = Tr(glyphs.power("{}", "3")); """Cubed value glyph."""
squared  = Tr(glyphs.power("{}", "2")); """Squared value glyph."""
transp   = Tr("{}.T"); """Matrix transposition glyph."""
ptransp  = Tr("{}.Tx"); """Matrix partial transposition glyph."""
htransp  = Tr("{}.H"); """Matrix hermitian transposition glyph."""

# Other operator glyphs.
add      = Op("{} + {}", 3, assoc=True);  """Addition glyph."""
sub      = Op("{} - {}", 3, assoc=False); """Substraction glyph."""
hadamard = Op("{}(o){}", 2, assoc=True);  """Hadamard product glyph."""
kron     = Op("{}(x){}", 2, assoc=True);  """Kronecker product glyph."""
mul      = Op("{}*{}",   2, assoc=True);  """Multiplication glyph."""
div      = Op("{}/{}",   2, assoc=False); """Division glyph."""
neg      = Op("-{}",     2.5);              """Negation glyph."""
plsmns   = Op("[+/-]{}", 2.5);              """Plus/Minus glyph."""

# Concatenation glyphs.
horicat  = Op("{}, {}", 4, assoc=True); """Horizontal concatenation glyph."""
vertcat  = Op("{}; {}", 4.5, assoc=True); """Vertical concatenation glyph."""

# Relation glyphs.
element  = Rl("{} in {}"); """Set element glyph."""
eq       = Rl("{} = {}");  """Equality glyph."""
ge       = Rl("{} >= {}"); """Greater or equal glyph."""
gt       = Rl("{} > {}");  """Greater than glyph."""
le       = Rl("{} <= {}"); """Lesser or equal glyph."""
lt       = Rl("{} < {}");  """Lesser than glyph."""
psdge    = Rl("{} >> {}"); """Lesser or equal w.r.t. the p.s.d. cone glyph."""
psdle    = Rl("{} << {}"); """Greater or equal w.r.t. the p.s.d. cone glyph."""


def latin1(rebuildDerivedGlyphs=True):
    """Let PICOS create future strings using only ISO 8859-1 characters."""
    # Reset to ASCII first.
    ascii()

    # Update glyphs with only template changes.
    glyphs.compose.template  = "{}°{}"
    glyphs.cubed.template    = "{}³"
    glyphs.hadamard.template = "{}(·){}"
    glyphs.kron.template     = "{}(×){}"
    glyphs.leadsto.template  = "{} » {}"
    glyphs.mul.template      = "{}·{}"
    glyphs.squared.template  = "{}²"
    glyphs.plsmns.template   = "±{}"
    glyphs.psdge.template    = "{} » {}"
    glyphs.psdle.template    = "{} « {}"
    glyphs.size.template     = "{}×{}"

    # Update all derived glyphs.
    if rebuildDerivedGlyphs:
        rebuild()


def unicode(rebuildDerivedGlyphs=True):
    """Let PICOS create future strings using only unicode characters."""
    # Reset to LATIN-1 first.
    latin1(rebuildDerivedGlyphs=False)

    # Update glyphs with only template changes.
    glyphs.and_.template     = "{} ∧ {}"
    glyphs.compose.template  = "{}∘{}"
    glyphs.dotp.template     = "⟨{}, {}⟩"
    glyphs.element.template  = "{} ∈ {}"
    glyphs.forall.template   = "{} ∀ {}"
    glyphs.fromto.template   = "{}…{}"
    glyphs.ge.template       = "{} ≥ {}"
    glyphs.hadamard.template = "{}⊙{}"
    glyphs.htransp.template  = "{}ᴴ"
    glyphs.infty.template    = "∞"
    glyphs.kron.template     = "{}⊗{}"
    glyphs.lambda_.template  = "λ"
    glyphs.le.template       = "{} ≤ {}"
    glyphs.leadsto.template  = "{} → {}"
    glyphs.norm.template     = "‖{}‖"
    glyphs.or_.template      = "{} ∨ {}"
    glyphs.psdge.template    = "{} ≽ {}"
    glyphs.psdle.template    = "{} ≼ {}"
    glyphs.sum.template      = "∑({})"
    glyphs.transp.template   = "{}ᵀ"

    # Update all derived glyphs.
    if rebuildDerivedGlyphs:
        rebuild()


# Default to unicode glyphs.
default = unicode
default()


# Helper functions that mimic or create additional glyphs.
# --------------------------------------------------------

def scalar(value):
    """Format a scalar value.

    This function mimics an operator glyph, but it returns a normal string (as
    opposed to an :class:`OpStr`) for nonnegative numbers.

    This is not realized as an atomic operator glyph to not increase the
    recursion depth of :func:`is_negated` and :func:`unnegate` unnecessarily.

    **Example**

    >>> from picos.glyphs import scalar
    >>> str(1.0)
    '1.0'
    >>> scalar(1.0)
    '1'
    """
    if not isinstance(value, complex) and value < 0:
        value = -value
        negated = True
    else:
        negated = False

    string = ("{:g}" if type(value) is float else "{}").format(value)

    if negated:
        return glyphs.neg(string)
    else:
        return string


def shape(theShape):
    """Describe a matrix shape that can contain wildcards.

    A wrapper around :obj:`size` that takes just one argument (the shape) that
    may contain wildcards which are printed as ``'?'``.
    """
    newShape = (
        "?" if theShape[0] is None else theShape[0],
        "?" if theShape[1] is None else theShape[1])
    return glyphs.size(*newShape)


def make_function(*names):
    """Create an ad-hoc composite function glyphs.

    **Example**

    >>> from picos.glyphs import make_function
    >>> make_function("log", "sum", "exp")("x")
    'log∘sum∘exp(x)'
    """
    return Fn("{}({{}})".format(functools.reduce(glyphs.compose, names)))


# Helper functions that make context-sensitive use of glyphs.
# -----------------------------------------------------------

def from_glyph(string, theGlyph):
    """Whether the given string was created by the given glyph."""
    return isinstance(string, GlStr) and string.glyph is theGlyph


CAN_FACTOR_OUT_NEGATION = (
    glyphs.matrix,
    glyphs.sum,
    glyphs.trace,
    glyphs.vec,
    glyphs.diag,
    glyphs.maindiag,
    glyphs.real
)  #: Operator glyphs for which negation may be factored out.


def is_negated(value):
    """Check if a value can be unnegated by :func:`unnegate`."""
    if isinstance(value, OpStr) and value.glyph in CAN_FACTOR_OUT_NEGATION:
        return is_negated(value.operands[0])
    elif from_glyph(value, glyphs.neg):
        return True
    elif type(value) is str:
        try:
            return float(value) < 0
        except ValueError:
            return False
    elif type(value) in (int, float):
        return value < 0
    else:
        return False


def unnegate(value):
    """Unnegate a value, usually a glyph-created string, in a sensible way.

    Unnegates a :class:`operator glyph created string <OpStr>` or other value in
    a sensible way, more precisely by recursing through a sequence of glyphs
    used to create the value and for which we can factor out negation, and
    negating the underlaying (numeric or string) value.

    :raises ValueError: When :meth:`is_negated` returns :obj:`False`.
    """
    if isinstance(value, OpStr) and value.glyph in CAN_FACTOR_OUT_NEGATION:
        return value.glyph(unnegate(value.operands[0]))
    elif from_glyph(value, glyphs.neg):
        return value.operands[0]
    elif type(value) is str:
        # We raise any conversion error, because is_negated returns False.
        return "{:g}".format(-float(value))
    elif type(value) in (int, float):
        return -value
    else:
        raise ValueError("The value to recursively unnegate is not negated in a"
            "supported manner.")


def clever_neg(value):
    """Describe the negation of a value in a clever way.

    A wrapper around :attr:`neg` that resorts to unnegating an already negated
    value.

    **Example**

    >>> from picos.glyphs import neg, clever_neg, matrix
    >>> neg("x")
    '-x'
    >>> neg(neg("x"))
    '-(-x)'
    >>> clever_neg(neg("x"))
    'x'
    >>> neg(matrix(-1))
    '-[-1]'
    >>> clever_neg(matrix(-1))
    '[1]'
    """
    if is_negated(value):
        return unnegate(value)
    else:
        return glyphs.neg(value)


def clever_add(left, right):
    """Describe the addition of two values in a clever way.

    A wrapper around :attr:`add` that resorts to :attr:`sub` if the second
    operand was created by :attr:`neg` or is a negative number (string). In both
    cases the second operand is adjusted accordingly.

    **Example**

    >>> from picos.glyphs import neg, add, clever_add, matrix
    >>> add("x", neg("y"))
    'x + -y'
    >>> clever_add("x", neg("y"))
    'x - y'
    >>> add("X", matrix(neg("y")))
    'X + [-y]'
    >>> clever_add("X", matrix(neg("y")))
    'X - [y]'
    >>> clever_add("X", matrix(-1.5))
    'X - [1.5]'
    """
    if left in (0, "0"):
        return right

    if right in (0, "0"):
        return left

    if is_negated(right):
        return glyphs.sub(left, unnegate(right))
    else:
        return glyphs.add(left, right)


def clever_sub(left, right):
    """Describe the substraction of a value from another in a clever way.

    A wrapper around :attr:`sub` that resorts to :attr:`add` if the second
    operand was created by :attr:`neg` or is a negative number(string). In both
    cases the second operand is adjusted accordingly.

    **Example**

    >>> from picos.glyphs import neg, sub, clever_sub, matrix
    >>> sub("x", neg("y"))
    'x - -y'
    >>> clever_sub("x", neg("y"))
    'x + y'
    >>> sub("X", matrix(neg("y")))
    'X - [-y]'
    >>> clever_sub("X", matrix(neg("y")))
    'X + [y]'
    >>> clever_sub("X", matrix(-1.5))
    'X + [1.5]'
    """
    if left in (0, "0"):
        return clever_neg(right)

    if right in (0, "0"):
        return left

    if is_negated(right):
        return glyphs.add(left, unnegate(right))
    else:
        return glyphs.sub(left, right)


def clever_mul(left, right):
    """Describe the multiplocation of two values in a clever way.

    A wrapper around :attr:`mul` that factors out identity factors.
    """
    # Apply a factor of zero on the left side.
    if left in (0, "0"):
        return left

    # Apply a factor of zero on the right side.
    if right in (0, "0"):
        return right

    # Detect quadratics.
    if left == right:
        return glyphs.squared(left)

    # Factor out a scalar one on the left side.
    if left in (1, "1"):
        return right

    # Factor out a scalar one on the right side.
    if right in (1, "1"):
        return left

    # Factor out negation.
    ln, rn = is_negated(left), is_negated(right)
    if ln and rn:
        return glyphs.clever_mul(unnegate(left), unnegate(right))
    elif ln:
        return glyphs.neg(glyphs.clever_mul(unnegate(left), right))
    elif rn:
        return glyphs.neg(glyphs.clever_mul(left, unnegate(right)))

    return glyphs.mul(left, right)


def clever_div(left, right):
    """Describe the division of one value by another in a clever way.

    A wrapper around :attr:`div` that factors out identity factors.
    """
    # Apply a factor of zero on the left side.
    if left in (0, "0"):
        return left

    # Factor out a scalar one on the right side.
    if right in (1, "1"):
        return left

    # Factor out negation.
    ln, rn = is_negated(left), is_negated(right)
    if ln and rn:
        return glyphs.clever_div(unnegate(left), unnegate(right))
    elif ln:
        return glyphs.neg(glyphs.clever_div(unnegate(left), right))
    elif rn:
        return glyphs.neg(glyphs.clever_div(left, unnegate(right)))

    return glyphs.div(left, right)


def clever_dotp(left, right, complexRHS, scalar=False):
    """Describe an inner product in a clever way.

    :param bool complexRHS: Whether the right hand side is complex.
    """
    riCo = glyphs.conj(right) if complexRHS else right

    if scalar:
        return glyphs.clever_mul(left, right)

    if   from_glyph(left, glyphs.idmatrix):
        return glyphs.trace(riCo)
    elif from_glyph(riCo, glyphs.idmatrix):
        return glyphs.trace(left)
    elif from_glyph(left, glyphs.matrix) and left.operands[0] in (1, "1"):
        return glyphs.sum(riCo)
    elif from_glyph(riCo, glyphs.matrix) and riCo.operands[0] in (1, "1"):
        return glyphs.sum(left)
    elif left == right:
        return glyphs.squared(glyphs.norm(left))
    else:
        return glyphs.dotp(left, right)


def matrix_cat(left, right, horizontal=True):
    """Describe matrix concatenation in a clever way.

    A wrapper around :attr:`matrix`, :attr:`horicat` and :attr:`vertcat`.

    **Example**

    >>> from picos.glyphs import matrix_cat
    >>> Z = matrix_cat("X", "Y")
    >>> Z
    '[X, Y]'
    >>> matrix_cat(Z, Z)
    '[X, Y, X, Y]'
    >>> matrix_cat(Z, Z, horizontal = False)
    '[X, Y; X, Y]'
    """
    if isinstance(left, OpStr) and left.glyph is glyphs.matrix:
        left = left.operands[0]

    if isinstance(right, OpStr) and right.glyph is glyphs.matrix:
        right = right.operands[0]

    catGlyph = glyphs.horicat if horizontal else glyphs.vertcat

    return glyphs.matrix(catGlyph(left, right))


def row_vectorize(*entries):
    """Describe a row vector with the given symbolic entries."""
    return functools.reduce(matrix_cat, entries)


def col_vectorize(*entries):
    """Describe a column vector with the given symbolic entries."""
    return functools.reduce(lambda l, r: matrix_cat(l, r, False), entries)


def free_var_name(string):
    """Return a variable name not present in the given string."""
    names = "xyzpqrstuvwabcdefghijklmno"
    for name in names:
        if name not in string:
            return name
    return "?"


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`SumExponentialsConstraint`."""

import math
from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class SumExponentialsConstraint(Constraint):
    """Upper bound on a sum of exponentials."""

    class ConicConversion(ConstraintConversion):
        """Sum of exponentials to exponential cone constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import RealVariable
            from . import AffineConstraint, ExpConeConstraint

            n = subtype.argdim

            yield ("var", RealVariable.make_var_type(dim=n, bnd=0), 1)
            yield ("con", AffineConstraint.make_type(dim=1, eq=False), 1)
            yield ("con", ExpConeConstraint.make_type(), n)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import ExponentialCone
            from ..modeling import Problem

            x = con.numerator
            y = con.denominator
            n = con.theSum.n
            b = con.upperBound

            P = Problem()

            u = P.add_variable("__u", n)
            P.add_constraint((u | 1) <= b)

            for i in range(n):
                P.add_constraint((u[i] // y[i] // x[i]) << ExponentialCone())

            return P

    class LogSumExpConversion(ConstraintConversion):
        """Sum of exponentials to logarithm of the sum constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from . import LogSumExpConstraint

            n = subtype.argdim

            if subtype.lse_representable:
                yield ("con", LogSumExpConstraint.make_type(argdim=n), 1)
            else:
                # HACK: Return the input constraint type.
                # TODO: Handle partial subtype support differently, e.g. by
                #       introducing ConstraintConversion.supports.
                yield ("con", SumExponentialsConstraint.make_type(*subtype), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import LogSumExp
            from ..modeling import Problem

            x = con.numerator
            b = con.upperBound

            P = Problem()

            if con.lse_representable:
                P.add_constraint(LogSumExp(x) <= math.log(b.value))
            else:
                # HACK: See predict.
                P.add_constraint(con)

            return P

    def __init__(self, theSum, upperBound):
        """Construct a :class:`SumExponentialsConstraint`.

        :param ~picos.expressions.SumExponentials theSum:
            Constrained expression.
        :param ~picos.expressions.AffineExpression upperBound:
            Upper bound on the expression.
        """
        from ..expressions import AffineExpression, SumExponentials

        assert isinstance(theSum, SumExponentials)
        assert isinstance(upperBound, AffineExpression)
        assert len(upperBound) == 1

        self.theSum     = theSum
        self.upperBound = upperBound

        super(SumExponentialsConstraint, self).__init__(theSum._typeStr)

    @property
    def numerator(self):
        """The :math:`x` of the sum."""
        return self.theSum.x

    @cached_property
    def denominator(self):
        """The :math:`y` of the sum, or :math:`1`."""
        if self.theSum.y is None:
            from ..expressions import AffineExpression
            return AffineExpression.from_constant(1, self.theSum.x.shape)
        else:
            return self.theSum.y

    @property
    def lse_representable(self):
        """Whether this can be converted to a logarithmic constraint."""
        if self.theSum.y is not None:
            return False

        if not self.upperBound.constant:
            return False

        if self.upperBound.value < 0:
            return False

        return True

    Subtype = namedtuple("Subtype", ("argdim", "lse_representable"))

    def _subtype(self):
        return self.Subtype(self.theSum.n, self.lse_representable)

    @classmethod
    def _cost(cls, subtype):
        # NOTE: Twice the argument dimension due to the denominator.
        return 2*subtype.argdim + 1

    def _expression_names(self):
        yield "theSum"
        yield "upperBound"

    def _str(self):
        return glyphs.le(self.theSum.string, self.upperBound.string)

    def _get_slack(self):
        return self.upperBound.value - self.theSum.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Optimization problem solution strategy search."""

from __future__ import print_function

from collections import OrderedDict

from .. import glyphs
from ..apidoc import api_end, api_start
from ..reforms import SORTED_REFORMS, ExtraOptions, Reformulation
from ..solvers import Solver, available_solvers, get_solver, get_solver_name
from .problem import Problem

_API_START = api_start(globals())
# -------------------------------


class NoStrategyFound(RuntimeError):
    """No solution strategy found.

    Raised when no viable combination of reformulations and solver to tackle the
    problem could be found.
    """

    def __str__(self):
        return super(NoStrategyFound, self).__str__() if self.args else \
            "PICOS failed to find a sequence of problem reformulations that " \
            "would allow an available solver to tackle the problem."


class Strategy:
    """Optimization problem solution strategy."""

    def __init__(self, problem, solver, *reforms):
        """Construct a :class:`Strategy`.

        :param ~picos.Problem problem:
            The first step in the solution pipeline; the problem to be solved.

        :param type solver:
            The last step in the solution pipeline; the solver class to be used.

        :param list(~picos.reforms.reformulation.Reformulation) reforms:
            Intermediate steps in the pipeline; reformulations to be applied.
            May not include :class:`~picos.reforms.ExtraOptions` which is
            automatically made the first reformulation.
        """
        if not isinstance(problem, Problem):
            raise TypeError("First argument must be a problem instance.")

        if not issubclass(solver, Solver):
            raise TypeError("Second argument must be a solver class.")

        if not all(issubclass(reform, Reformulation) for reform in reforms):
            raise TypeError("Extra arguments must be reformulation classes.")

        if ExtraOptions in reforms:
            raise TypeError("The ExtraOptions reformulation is implicitly part "
                "of any strategy and may not be added explicitly.")

        self.nodes = [problem, ExtraOptions(problem)]

        for reform in reforms:
            self.nodes.append(reform(self.nodes[-1]))

        self.nodes.append(solver(self.nodes[-1]))

    @property
    def problem(self):
        """The problem to be solved."""
        return self.nodes[0]

    @property
    def reforms(self):
        """All reformulations in use.

        This includes the implicit :class:`~picos.reforms.ExtraOptions`.
        """
        return self.nodes[1:-1]

    @property
    def solver(self):
        """The solver instance in use."""
        return self.nodes[-1]

    def __str__(self):
        return "\n".join(
            "{}. {}".format(num + 1, node.__class__.__name__)
            for num, node in enumerate(self.nodes[1:]))

    def __repr__(self):
        return glyphs.repr1("Solution strategy for {}".format(self.solver.name))

    def valid(self, **extra_options):
        """Whether the solution strategy can be executed.

        :param extra_options:
            A keyword parameter sequence of additional options (in addition to
            those of the problem) to assume used.
        """
        problem = self.nodes[0]
        solver = self.nodes[-1]

        # Determine the footprint with extra options set.
        footprint = problem.footprint.with_extra_options(**extra_options)
        options = footprint.options

        # Handle a conflicting solver selection.
        if options.solver and options.solver != get_solver_name(solver):
            return False

        # Skip ExtraOptions but include the solver with the following.
        for node in self.nodes[2:]:
            if not node.supports(footprint):
                return False

            footprint = node.predict(footprint)

        return True

    def execute(self, **extra_options):
        """Execute the solution strategy.

        :param extra_options:
            A keyword parameter sequence of additional options (in addition to
            those of the problem) to use for this search.

        :returns:
             :class:`~picos.modeling.Solution` to the problem.
        """
        # Defer solving to the first reformulation, which is responsible for
        # applying the extra options (i.e. ExtraOptions).
        solution = self.nodes[1].execute(**extra_options)

        # Attach the solution to the root problem. Note that reformulations are
        # allowed to already do this within their 'backward' method, but for
        # performance reasons it is best to do this just once, here.
        if isinstance(solution, list):
            for s in solution:
                s.attach_to(self.nodes[0])
        else:
            solution.attach_to(self.nodes[0])

        return solution

    @classmethod
    def from_problem(cls, problem, **extra_options):
        """Create a solution strategy for the given problem.

        :param ~picos.Problem problem:
            The optimization problem to search a strategy for.

        :param extra_options:
            A keyword parameter sequence of additional options (in addition to
            those of the problem) to assume used.
        """
        # Determine the footprint with extra options set.
        footprint = problem.footprint.with_extra_options(**extra_options)
        options = footprint.options

        # Decide on solvers to consider.
        solvers = []
        if options.solver:
            solver = get_solver(options.solver)

            if solver.available():
                solvers.append(solver)
            else:
                raise RuntimeError(
                    "Selected solver {} is not available on the system."
                    .format(solver.names()[1]))
        else:
            for solver_name in available_solvers():
                solver = get_solver(solver_name)
                solvers.append(solver)

        assert solvers, "Not even CVXOPT seems to be available."

        if len(solvers) == 1 and solvers[0].supports(footprint):
            if options.verbosity >= 2:
                print("{} supports the problem directly.".format(
                    solvers[0].names()[1]))

            return cls(problem, solvers[0])

        if options.verbosity >= 2:
            print("Selected solvers:\n  {}".format(", ".join(
                solver.names()[1] for solver in solvers)))

        paths = OrderedDict({footprint: tuple()})
        new_footprints = [footprint]

        while new_footprints:
            if options.max_footprints is not None \
            and len(paths) >= options.max_footprints:
                if options.verbosity >= 1:
                    print("Footprint limit reached ({}/{}).".format(
                        len(paths), options.max_footprints))
                break

            active_footprints = new_footprints
            new_footprints = []

            if options.verbosity >= 3:
                print("Active footprints:\n{}".format("\n".join("  ({}) {}"
                    .format(len(paths) - len(active_footprints) + i, f)
                    for i, f in enumerate(active_footprints))))

            for num, footprint in enumerate(active_footprints):
                if options.verbosity >= 3:
                    print("Prediction for ({}):".format(
                        len(paths) - len(active_footprints) + num))

                for step in SORTED_REFORMS:
                    if options.verbosity >= 3:
                        print("  {}:".format(step.__name__), end=" ")

                    # Don't apply the same reformulation multiple times.
                    if step in paths[footprint]:
                        if options.verbosity >= 3:
                            print("Already part of current path.")
                        continue

                    # Check if the reformulation applies.
                    if not step.supports(footprint):
                        if options.verbosity >= 3:
                            print("Not supported.")
                        continue

                    # Predict the reformulation outcome.
                    new_footprint = step.predict(footprint)

                    # Remember only the first (shortest) path to any footprint.
                    if new_footprint in paths:
                        if options.verbosity >= 3:
                            print("Resulting footprint already reached.")
                        continue

                    if options.verbosity >= 3:
                        print("Reached new footprint ({}).".format(len(paths)))

                    paths[new_footprint] = paths[footprint] + (step,)
                    new_footprints.append(new_footprint)

        # Sort footprints by cost. 'paths' being an ordered dict ensures a
        # deterministic order with respect to same-cost footprints (with shorter
        # reformulation pipelines taking precedence).
        footprints = sorted(paths, key=(lambda f: f.cost))

        if options.verbosity >= 2:
            print("Reachable footprints:\n{}".format("\n".join(
                "  ({}) [{:.4f}] {}".format(i, f.cost, f)
                for i, f in enumerate(footprints))))

        strategies = []
        costs = []

        if options.verbosity >= 2:
            print("Solvable footprints:")

        for num, footprint in enumerate(footprints):
            if not solvers:
                break

            for solver in list(solvers):
                if solver.supports(footprint):
                    cost = footprint.cost
                    penalty = solver.penalty(options)
                    total_cost = cost + penalty

                    if options.verbosity >= 2:
                        print("  {} supports ({}) at cost {:.2f} + {:.2f} = "
                            "{:.2f}.".format(solver.names()[1], num, cost,
                            penalty, total_cost))

                    strategies.append(cls(problem, solver, *paths[footprint]))
                    costs.append(total_cost)

                    solvers.remove(solver)

        if not strategies:
            if options.verbosity >= 2:
                print("  None found.")

            raise NoStrategyFound

        return strategies[costs.index(min(costs))]


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Backend for constraint type implementations."""

import random
import threading

import cvxopt

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property, empty_cache
from ..compat import ABC, abstractmethod
from ..containers import DetailedType

_API_START = api_start(globals())
# -------------------------------


# The constraint IDs start at a random value to prevent a clash if constraints
# are are pickled and loaded in another python session.
_LAST_CONSTRAINT_ID = int(random.getrandbits(31))

# A lock for _LAST_CONSTRAINT_ID, if the user threads to create constraints.
_CONSTRAINT_ID_LOCK = threading.Lock()


def _make_constraint_id():
    """Create a unique ID for a new constraint."""
    with _CONSTRAINT_ID_LOCK:
        global _LAST_CONSTRAINT_ID
        _LAST_CONSTRAINT_ID += 1
        return _LAST_CONSTRAINT_ID


class ConstraintType(DetailedType):
    """Container for a pair of constraint class type and constraint subtype."""

    pass


class Constraint(ABC):
    """Abstract base class for optimization constraints.

    Implementations

        * need to implement at least the abstract methods ``_str``,
          ``_expression_names``, and ``_get_slack``,
        * need to implement ``_get_size``, unless duals are not supported, and
        * are supposed to call :meth:`Constraint.__init__` from within their own
          implementation of ``__init__``.
    """

    LE = "<"
    GE = ">"
    EQ = "="

    def __init__(self, typeTerm, customString=None, printSize=False):
        """Perform basic initialization for :class:`Constraint` instances.

        :param str typeTerm: Short string denoting the constraint type.
        :param str customString: Optional string description.
        :param bool printSize: Whether to include the constraint's shape in its
            representation string.
        """
        self.typeTerm     = typeTerm
        self.customString = customString
        self.printSize    = printSize

        self._id = _make_constraint_id()

        self.name  = None
        self._dual = None

    @property
    def type(self):
        """Detailed type of the constraint.

        The detailed type of the constraint, which is suffcient to predict the
        outcome (detailed types and quantities of auxilary variables and
        constraints) of any constraint conversion.
        """
        return ConstraintType(self.__class__, self._subtype())

    subtype = property(lambda self: self._subtype())

    @classmethod
    def make_type(cls, *args, **kwargs):
        """Create a detailed constraint type from subtype parameters."""
        return ConstraintType(cls, cls.Subtype(*args, **kwargs))

    @abstractmethod
    def _subtype(self):
        """Subtype of the constraint.

        :returns: A hashable object that, together with the constraint class,
            is sufficient to predict the outcome (detailed types and quantities
            of auxilary variables and constraints) of any constraint conversion.
        """
        pass

    @classmethod
    @abstractmethod
    def _cost(cls, subtype):
        r"""Report an estimate number of real constraint matrix rows occupied.

        Given the subtype of a constraint, returns an estimate on the number of
        rows that a constraint with this subtype would occupy in the constraint
        matrix of a (hypothetical) solver that supports direct input of such a
        constraint.

        Some conventions:

        - For a conic constraint, this is the cone's dimensionality.
        - For a (complex) affine equality :math:`A = B`, this is the number of
          elements in the matrix :math:`A - B` (times two).
        - For a constraint that poses a bound on a scalar function on an
          :math:`n`-dimensional vector space, this is :math:`n + 1` (:math:`1`
          for the affine bound).
        - In particular, a bound on a function of symmetric (hermitian) matrices
          occupies :math:`\frac{n(n + 1)}{2} + 1` (:math:`n^2 + 1`) rows.
        - For a quadratic constraint, this is the number of coefficients in the
          simplified quadratic form, plus one (for the affine part).
        """
        pass

    def __hash__(self):
        """Return the unique ID."""
        return self._id

    def __eq__(self, other):
        """Whether the unique IDs equal."""
        return self._id == other._id

    def __repr__(self):
        if self.printSize:
            return glyphs.repr2("{} {} Constraint".format(
                glyphs.size(*self.size), self.typeTerm), self.__str__())
        else:
            return glyphs.repr2("{} Constraint".format(self.typeTerm),
                self.__str__())

    def __str__(self):
        return "{}{}".format(
            self.customString if self.customString else self._str(),
            " ({})".format(self.name) if self.name else "")

    def __len__(self):
        """Return number of scalar Lagrange dual variables."""
        return self.size[0] * self.size[1]

    @property
    def id(self):
        """The unique ID of the constraint, assigned at creation.

        The ID is kept when the constraint is copied via
        :meth:`replace_variables`, so that the copy can be identified with the
        original despite pointing to different expressions and variable objects.
        """
        return self._id

    @abstractmethod
    def _expression_names(self):
        """Attribute names of the expressions stored in the constraint."""
        pass

    @abstractmethod
    def _str(self):
        """Algebraic representation of the constraint."""
        pass

    def _get_size(self):
        """Langrange dual variable shape.

        The dimensionality of the constraint, more precisely the dimensionality
        of its Lagrange dual variable, as a pair.
        """
        raise NotImplementedError(
            "{} does not define a constraint (dual value) dimensionality."
            .format(self.__class__.__name__))

    @abstractmethod
    def _get_slack(self):
        """Value of a slack variable or of the negative constraint violation.

        A negative value whose absolute value corresponds to the amount of
        violation, if the constraint is violated, or a non-negative value that
        corresponds to the value of a slack variable, otherwise.
        """
        pass

    def _wrap_get_slack(self):
        """Convert a scalar slack value to float.

        A wrapper retrieving the slack in a consistent manner: If it is a scalar
        value, it is returned as a float, otherwise as a
        :class:`CVXOPT matrix <cvxopt.matrix>`.
        """
        slack = self._get_slack()

        if isinstance(slack, float):
            return slack

        assert isinstance(slack, cvxopt.matrix), "Constraints need to return " \
            "the slack as either a float or a CVXOPT matrix."

        if slack.size == (1, 1):
            return float(slack[0])
        else:
            return slack

    def _get_dual(self):
        return self._dual

    def _set_dual(self, value):
        """Store a constraint's dual value.

        Stores a dual solution value for the dual variable corresponding to the
        constraint in a consistent manner.

        Duals for multidmensional constraints are stored as a
        :class:`CVXOPT matrix <cvxopt.matrix>` while duals for scalar
        constraints are stored as a float.
        """
        if value is None:
            self._dual = None
        elif type(value) in (int, float, complex):
            if len(self) != 1:
                raise ValueError("Incompatible dimensions of dual value: "
                    "Expected {} but got a scalar.".format(self.size))

            self._dual = value
        else:
            try:
                dual = cvxopt.matrix(value, self.size)
            except TypeError:
                raise ValueError("Incompatible dimensions of dual value for {}:"
                    " Expected {} and got container of size {}."
                    .format(self.__class__.__name__, glyphs.size(*self.size),
                    len(value)))

            self._dual = dual[0] if len(dual) == 1 else dual

    size  = property(lambda self: self._get_size(), doc=_get_size.__doc__)
    slack = property(lambda self: self._wrap_get_slack(),
        doc=_get_slack.__doc__)
    dual  = property(lambda self: self._get_dual(), _set_dual)
    """Value of the constraint's Lagrange dual variable."""

    def delete(self):
        """Raise a :exc:`NotImplementedError`.

        Formerly this would remove the constraint from the single problem it is
        assigned to, if any.

        .. deprecated:: 2.0

            Both variables and constraints have been decoupled from problems:
            Both may safely appear in multiple problems but at the same time
            they do not know which problems they were added to. To remove a
            constraint from a problem, you have to call its
            :meth:`~picos.modeling.problem.Problem.remove_constraint` method.
        """
        raise NotImplementedError("Telling a constraint to remove itself from "
            "problems is not possible anymore, use Problem.remove_constraint "
            "on selected problems instead.")

    # TODO: Solve this with inspection instead of _expression_names?
    # TODO: Is this even needed for anything apart from variables and
    #       replace_variables? If not, maybe just implement those two with
    #       every constraint just like for expressions?
    @property
    def expressions(self):
        """Yield expressions stored with the constraint."""
        for name in self._expression_names():
            yield getattr(self, name)

    @cached_property
    def variables(self):
        """All variables referenced by the constraint."""
        vars = frozenset()
        for expression in self.expressions:
            vars = vars.union(expression.variables)
        return vars

    def replace_variables(self, new_variables):
        """Make the constraint concern a different set of variables.

        See :meth:`~.expression.Expression.replace_variables` for more.
        """
        from copy import copy

        theCopy = copy(self)

        # TODO: Reset any dual value assigned to the constraint?
        # theCopy._dual = None

        # Clear the cache of the copy as it can reference old variables.
        empty_cache(theCopy)

        # Rewrite expressions.
        for name in theCopy._expression_names():
            expression = getattr(self, name)
            new_expression = expression.replace_variables(new_variables)
            setattr(theCopy, name, new_expression)

        # HACK: Delete a custom string because it can contain old variable
        #       names. In particular Norm uses it when creating a SOCConstraint.
        # TODO: Get rid of custom strings.
        theCopy.customString = None

        return theCopy

    # TODO: Evaluate uses of this method.
    def constring(self):
        """Return an algebraic string representation of the constraint."""
        return self._str()

    # TODO: Re-implement pretty printing for problems.
    def keyconstring(self):
        """Return the regular string representation."""
        return self.__str__()

    def _assure_lhs_rhs_relation(self):
        if not hasattr(self, "relation") or not hasattr(self, "lhs") \
        or not hasattr(self, "rhs"):
            raise TypeError("{} does not explicitly define a relation "
                "between a left hand side and a right hand side expression."
                .format(self.__class__.__name__))

    def is_equality(self):
        """Whether the constraints states an equality."""
        self._assure_lhs_rhs_relation()
        return self.relation == self.EQ

    def is_inequality(self):
        """Whether the constraints states an inequality."""
        self._assure_lhs_rhs_relation()
        return self.relation != self.EQ

    def is_increasing(self):
        """Whether the left side is posed smaller or equal than the right."""
        self._assure_lhs_rhs_relation()
        return self.relation == self.LE

    def is_decreasing(self):
        """Whether the left side is posed greater or equal than the right."""
        self._assure_lhs_rhs_relation()
        return self.relation == self.GE


class ConstraintConversion(ABC):
    """Recipe for conversion from one constraint to a set of other constraints.

    Implementations of this class are defined within the class body of a
    Constraint implementation to tell PICOS' reformulation framework how that
    constraint can be reformulated into a number of other constraints and
    auxiliary variables.

    Implementation class names must end in ``Conversion``, and in particular may
    be called just ``Conversion``. If for instance
    :class:`AbsoluteValueConstraint <picos.constraints.AbsoluteValueConstraint>`
    defines :class:`AffineConversion
    <picos.constraints.AbsoluteValueConstraint.AffineConversion>`, then the
    reformulation will be coined ``AbsoluteValueToAffineReformulation``. If the
    conversions was just named ``Conversion``, the result would be a class named
    ``AbsoluteValueReformulation``.
    """

    @classmethod
    @abstractmethod
    def predict(cls, subtype, options):
        """Predict the outcome of a constraint conversion.

        :param object subtype: A hashable object as could be returned by the
            ``_subtype`` method of the parent constraint implementation.
        :param ~picos.Options options: Solution options to assume used.
        :yields: Records to be added to a problem footprint when an instance of
            the parent constraint with the given subtype is converted according
            to this conversion.
        """
        pass

    # TODO: Make this yield variables and constraints instad of returning a
    #       Problem instance (this was not possible with the old expressions).
    @classmethod
    @abstractmethod
    def convert(cls, constraint, options):
        """Convert a given constraint.

        Returns a temporary problem instance that contains auxilary constraints
        and variables replacing the given constraint.
        """
        pass

    @classmethod
    def dual(cls, auxVarPrimals, auxConDuals, options):
        """Convert back the dual value of a constraint that was converted.

        Given a mapping of auxilary variable names (as named in :meth:`convert`)
        to primals and a list of auxilary constraint duals (in the order as the
        constraints were added in :meth:`convert`), returns a dual value for the
        converted constraint.

        :raises NotImplementedError: When dual format not decided upon or not
            known. This will be caught by the reformulation's backward method.
        """
        raise NotImplementedError(
            "{} does not describe how to convert back the dual.".format(
            cls.__qualname__ if hasattr(cls, "__qualname__") else cls.__name__))

    def __init__(self):
        """Raise a :exc:`TypeError` on instanciation."""
        raise TypeError("Constraint conversion classes are not supposed to be "
            "instanciated.")


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2017-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`MOSEKFusionSolver`."""

import sys

import cvxopt

from ..apidoc import api_end, api_start
from ..compat import assert_import_exists
from ..constraints import (AffineConstraint, LMIConstraint, RSOCConstraint,
                           SOCConstraint)
from ..expressions import AffineExpression, BinaryVariable, IntegerVariable
from ..modeling.footprint import Specification
from ..modeling.solution import (PS_FEASIBLE, PS_ILLPOSED, PS_INF_OR_UNB,
                                 PS_INFEASIBLE, PS_UNBOUNDED, PS_UNKNOWN,
                                 SS_EMPTY, SS_FEASIBLE, SS_INFEASIBLE,
                                 SS_OPTIMAL, SS_UNKNOWN)
from .solver import ProblemUpdateError, Solver

_API_START = api_start(globals())
# -------------------------------


class MOSEKFusionSolver(Solver):
    """Interface to the MOSEK solver via its high level Fusion API.

    Supports both MOSEK 8 and 9.

    The Fusion API is currently much slower than MOSEK's low level Python API.
    If this changes in the future, the Fusion API would be the prefered
    interface.
    """

    SUPPORTED = Specification.compile(
        directions=Specification.ALL,
        objectives=[
            AffineExpression],
        variables=Specification.ALL,
        constraints=[
            AffineConstraint,
            SOCConstraint,
            RSOCConstraint,
            LMIConstraint])

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.solver.Solver.supports`."""
        if not Solver.supports(footprint):
            return False

        # No integer SDPs.
        if footprint.integer and ("con", LMIConstraint) in footprint:
            return False

        return footprint << cls.SUPPORTED

    @classmethod
    def default_penalty(cls):
        """Implement :meth:`~.solver.Solver.default_penalty`."""
        return 0.5  # Commercial solver with slower interface.

    @classmethod
    def test_availability(cls):
        """Implement :meth:`~.solver.Solver.test_availability`."""
        assert_import_exists("mosek.fusion")

    @classmethod
    def names(cls):
        """Implement :meth:`~.solver.Solver.names`."""
        return "mskfsn", "MOSEK (Fusion)", "MOSEK via Fusion API"

    def __init__(self, problem):
        """Initialize a MOSEK (Fusion) solver interface.

        :param ~picos.Problem problem: The problem to be solved.
        """
        super(MOSEKFusionSolver, self).__init__(problem)

        # Maps PICOS variables to MOSEK variables and vice versa.
        self.knownVariables = {}

        # Maps PICOS constraints to MOSEK constraints and vice versa.
        self.knownConstraints = {}

    def __del__(self):
        if self.int is not None:
            self.int.dispose()

    def reset_problem(self):
        """Implement :meth:`~.solver.Solver.reset_problem`."""
        if self.int is not None:
            self.int.dispose()
        self.int = None
        self.knownVariables.clear()
        self.knownConstraints.clear()

    @classmethod
    def _get_major_version(cls):
        if not hasattr(cls, "mosekVersion"):
            import mosek
            cls.mosekVersion = mosek.Env.getversion()

        return cls.mosekVersion[0]

    ver = property(lambda self: self.__class__._get_major_version())
    """The major version of the available MOSEK library."""

    @classmethod
    def _mosek_sparse_triple(cls, I, J, V):
        """Transform a sparse triple (e.g. from CVXOPT) for use with MOSEK."""
        if cls._get_major_version() >= 9:
            IJV = list(IJV for IJV in zip(I, J, V) if IJV[2] != 0)
            I, J, V = (list(X) for X in zip(*IJV)) if IJV else ([], [], [])
        else:
            I = list(I) if not isinstance(I, list) else I
            J = list(J) if not isinstance(J, list) else J
            V = list(V) if not isinstance(V, list) else V

        return I, J, V

    @classmethod
    def _matrix_cvx2msk(cls, cvxoptMatrix):
        """Transform a CVXOPT (sparse) matrix into a MOSEK (sparse) matrix."""
        import mosek.fusion as msk

        M = cvxoptMatrix
        n, m = M.size

        if type(M) is cvxopt.spmatrix:
            return msk.Matrix.sparse(
                n, m, *cls._mosek_sparse_triple(M.I, M.J, M.V))
        elif type(M) is cvxopt.matrix:
            return msk.Matrix.dense(n, m, list(M.T))
        else:
            raise ValueError("Argument must be a CVXOPT matrix.")

    @classmethod
    def _mosek_vstack(cls, *expressions):
        """Vertically stack MOSEK expressions.

        This is a wrapper around MOSEK's :func:`vstack
        <mosek.fusion.Expr.vstack>` function that silences a FutureWarning.
        """
        import mosek.fusion as msk

        if cls._get_major_version() >= 9:
            return msk.Expr.vstack(*expressions)
        else:
            import warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", FutureWarning)
                return msk.Expr.vstack(*expressions)

    def _affinexp_pic2msk(self, picosExpression):
        """Transform an affine expression from PICOS to MOSEK.

        Requries all contained variables to be known to MOSEK.
        """
        import mosek.fusion as msk

        assert isinstance(picosExpression, AffineExpression)

        vectorShape = [len(picosExpression), 1]
        targetShape = list(picosExpression.size)

        if self.ver < 9:
            vectorShape = msk.Set.make(vectorShape)
            targetShape = msk.Set.make(targetShape)

        # Convert linear part of expression.
        firstSummand = True
        for picosVar, factor in picosExpression._coefs.items():
            mosekVar = self.knownVariables[picosVar]

            summand = msk.Expr.mul(self._matrix_cvx2msk(factor), mosekVar)
            if firstSummand:
                mosekExpression = summand
                firstSummand = False
            else:
                mosekExpression = msk.Expr.add(mosekExpression, summand)

        # Convert constant term of expression.
        if picosExpression.constant is not None:
            mosekConstant = msk.Expr.constTerm(
                self._matrix_cvx2msk(picosExpression._const))

            if firstSummand:
                mosekExpression = mosekConstant
            else:
                mosekExpression = msk.Expr.add(mosekExpression, mosekConstant)
        elif firstSummand:
            mosekExpression = msk.Expr.zeros(vectorShape)

        # Restore the expression's original shape.
        # NOTE: Transposition due to differing major orders.
        mosekExpression = msk.Expr.reshape(
            msk.Expr.transpose(mosekExpression), targetShape)

        if self._debug():
            self._debug(
                "Affine expression converted: {} → {}".format(
                repr(picosExpression), mosekExpression.toString()))

        return mosekExpression

    @classmethod
    def _bounds_pic2msk(cls, picosVar, fixMOSEK9=False):
        """Transform PICOS variable bounds to MOSEK matrices or scalars.

        Scalars are returned in the case of homogenous bounds.
        """
        import mosek.fusion as msk

        dim = picosVar.dim
        lower, upper = picosVar.bound_dicts

        if fixMOSEK9:
            LV, LI, LJ = [-1e20]*dim, list(range(dim)), [0]*dim
            UV, UI, UJ = [+1e20]*dim, list(range(dim)), [0]*dim

            for i, b in lower.items():
                LV[i] = b

            for i, b in upper.items():
                UV[i] = b
        else:
            LV, LI = [], []
            UV, UI = [], []

            for i, b in lower.items():
                LI.append(i)
                LV.append(b)

            for i, b in upper.items():
                UI.append(i)
                UV.append(b)

            LJ = [0]*len(LV)
            UJ = [0]*len(UV)

        mosekBounds = [None, None]
        for side, I, J, V in ((0, LI, LJ, LV), (1, UI, UJ, UV)):
            if len(V) == dim and len(set(V)) == 1:
                mosekBounds[side] = V[0]
            elif V:
                mosekBounds[side] = msk.Matrix.sparse(dim, 1, I, J, V)

        return mosekBounds

    def _import_variable(self, picosVar):
        import mosek.fusion as msk

        shape = [picosVar.dim, 1]

        # Import variable bounds.
        if not isinstance(picosVar, BinaryVariable):
            # Retrieve lower and upper bounds.
            lower, upper = self._bounds_pic2msk(picosVar)

            # Convert bounds to a domain.
            if lower is None and upper is None:
                domain = msk.Domain.unbounded()
            elif lower is not None and upper is None:
                domain = msk.Domain.greaterThan(lower)
            elif lower is None and upper is not None:
                domain = msk.Domain.lessThan(upper)
            elif lower is not None and upper is not None:
                if lower == upper:
                    domain = msk.Domain.equalsTo(lower)
                elif self.ver >= 9:
                    # HACK: MOSEK 9 does not accept sparse (partial) range
                    #       domains anymore. The workaround triggers a MOSEK
                    #       warning, but there is no other way to pass such
                    #       variable bounds directly.
                    if isinstance(lower, msk.Matrix) \
                    or isinstance(upper, msk.Matrix):
                        lower, upper = self._bounds_pic2msk(picosVar, True)
                        if isinstance(lower, msk.Matrix):
                            lower = lower.getDataAsArray()
                        if isinstance(upper, msk.Matrix):
                            upper = upper.getDataAsArray()

                    domain = msk.Domain.inRange(lower, upper, shape)
                else:
                    domain = msk.Domain.inRange(lower, upper)

        # Refine the domain with the variable's type.
        if isinstance(picosVar, BinaryVariable):
            domain = msk.Domain.binary()
        elif isinstance(picosVar, IntegerVariable):
            domain = msk.Domain.integral(domain)

        # Create the MOSEK variable.
        mosekVar = self.int.variable(picosVar.name, shape, domain)

        # Map the PICOS variable to the MOSEK variable and vice versa.
        self.knownVariables[picosVar] = mosekVar
        self.knownVariables[mosekVar] = picosVar

        if self._debug():
            self._debug("Variable imported: {} → {}"
                .format(picosVar, " ".join(mosekVar.toString().split())))

    # TODO: This needs a test.
    def _import_variable_values(self, integralOnly=False):
        for picosVar in self.ext.variables.values():
            if integralOnly and not isinstance(
                    picosVar, (BinaryVariable, IntegerVariable)):
                continue

            if picosVar.valued:
                value = picosVar.internal_value

                if isinstance(value, cvxopt.spmatrix):
                    value = cvxopt.matrix(value)

                self.knownVariables[picosVar].setLevel(list(value))

    def _import_linear_constraint(self, picosCon):
        import mosek.fusion as msk

        assert isinstance(picosCon, AffineConstraint)

        # Separate constraint into a linear function and a constant.
        linear, bound = picosCon.bounded_linear_form()

        # Rewrite constraint in MOSEK types: The linear function is represented
        # as a MOSEK expression while the constant term becomes a MOSEK domain.
        linear = self._affinexp_pic2msk(linear[:])
        bound  = self._matrix_cvx2msk(bound._const)

        if picosCon.is_increasing():
            domain = msk.Domain.lessThan(bound)
        elif picosCon.is_decreasing():
            domain = msk.Domain.greaterThan(bound)
        elif picosCon.is_equality():
            domain = msk.Domain.equalsTo(bound)
        else:
            assert False, "Unexpected constraint relation."

        # Import the constraint.
        if picosCon.name is None:
            return self.int.constraint(linear, domain)
        else:
            return self.int.constraint(picosCon.name, linear, domain)

    def _import_socone_constraint(self, picosCon):
        import mosek.fusion as msk

        assert isinstance(picosCon, SOCConstraint)

        coneElement = self._mosek_vstack(
            msk.Expr.flatten(self._affinexp_pic2msk(picosCon.ub)),
            msk.Expr.flatten(self._affinexp_pic2msk(picosCon.ne)))

        # TODO: Remove zeros from coneElement[1:].

        return self.int.constraint(coneElement, msk.Domain.inQCone())

    def _import_rscone_constraint(self, picosCon):
        import mosek.fusion as msk

        assert isinstance(picosCon, RSOCConstraint)

        # MOSEK handles the vector [x₁; x₂; x₃] as input for a constraint of the
        # form ‖x₃‖² ≤ 2x₁x₂ whereas PICOS handles the expressions e₁, e₂ and e₃
        # for a constraint of the form ‖e₁‖² ≤ e₂e₃.
        # Neutralize MOSEK's additional factor of two by scaling e₂ and e₃ by
        # sqrt(0.5) each to obtain x₁ and x₂ respectively.
        scale = 0.5**0.5
        coneElement = self._mosek_vstack(
            msk.Expr.flatten(self._affinexp_pic2msk(scale * picosCon.ub1)),
            msk.Expr.flatten(self._affinexp_pic2msk(scale * picosCon.ub2)),
            msk.Expr.flatten(self._affinexp_pic2msk(picosCon.ne)))

        # TODO: Remove zeros from coneElement[2:].

        return self.int.constraint(coneElement, msk.Domain.inRotatedQCone())

    def _import_sdp_constraint(self, picosCon):
        import mosek.fusion as msk
        assert isinstance(picosCon, LMIConstraint)

        semiDefMatrix = self._affinexp_pic2msk(picosCon.psd)

        return self.int.constraint(semiDefMatrix, msk.Domain.inPSDCone())

    def _import_constraint(self, picosCon):
        import mosek.fusion as msk

        # HACK: Work around faulty MOSEK warnings (warning 705).
        import os
        with open(os.devnull, "w") as devnull:
            self.int.setLogHandler(devnull)

            if isinstance(picosCon, AffineConstraint):
                mosekCon = self._import_linear_constraint(picosCon)
            elif isinstance(picosCon, SOCConstraint):
                mosekCon = self._import_socone_constraint(picosCon)
            elif isinstance(picosCon, RSOCConstraint):
                mosekCon = self._import_rscone_constraint(picosCon)
            elif isinstance(picosCon, LMIConstraint):
                mosekCon = self._import_sdp_constraint(picosCon)
            else:
                assert False, "Constraint type not supported."

            self.int.setLogHandler(sys.stdout)

        # Map the PICOS constraint to the MOSEK constraint and vice versa.
        self.knownConstraints[picosCon] = mosekCon
        self.knownConstraints[mosekCon] = picosCon

        if self._debug():
            self._debug("Constraint imported: {} → {}".format(picosCon,
                " ".join(mosekCon.toString().split()) if not isinstance(
                mosekCon, msk.PSDConstraint) else mosekCon))

    def _import_objective(self):
        import mosek.fusion as msk

        picosSense, picosObjective = self.ext.no

        if picosSense == "min":
            mosekSense = msk.ObjectiveSense.Minimize
        else:
            assert picosSense == "max"
            mosekSense = msk.ObjectiveSense.Maximize

        mosekObjective = self._affinexp_pic2msk(picosObjective)

        self.int.objective(mosekSense, mosekObjective)

        if self._debug():
            self._debug(
                "Objective imported: {} {} → {} {}".format(
                picosSense, picosObjective, mosekSense,
                " ".join(mosekObjective.toString().split())))

    def _import_problem(self):
        import mosek.fusion as msk

        # Create a problem instance.
        self.int = msk.Model()
        self.int.setLogHandler(sys.stdout)

        # Import variables.
        for variable in self.ext.variables.values():
            self._import_variable(variable)

        # Import constraints.
        for constraint in self.ext.constraints.values():
            self._import_constraint(constraint)

        # Set objective.
        self._import_objective()

    def _update_problem(self):
        for oldConstraint in self._removed_constraints():
            raise ProblemUpdateError(
                "MOSEK does not support removal of constraints.")

        for oldVariable in self._removed_variables():
            raise ProblemUpdateError(
                "MOSEK does not support removal of variables.")

        for newVariable in self._new_variables():
            self._import_variable(newVariable)

        for newConstraint in self._new_constraints():
            self._import_constraint(newConstraint)

        if self._objective_has_changed():
            self._import_objective()

    def _solve(self):
        import mosek.fusion as msk
        from mosek import objsense

        # MOSEK 8 has additional parameters and status codes.
        mosek8 = self.ver < 9

        # Reset options.
        # HACK: This is a direct access to MOSEK's internal Task object, which
        #       is necessary as the Fusion API has no call to reset options.
        # TODO: As soon as the Fusion API offers option reset, use it instead.
        self.int.getTask().setdefaults()

        # verbosity
        self.int.setSolverParam("log", max(0, self.verbosity()))

        # abs_prim_fsb_tol
        if self.ext.options.abs_prim_fsb_tol is not None:
            value = self.ext.options.abs_prim_fsb_tol

            # Interior-point primal feasibility tolerances.
            for ptype in ("", "Co") + (("Qo",) if mosek8 else ()):
                self.int.setSolverParam("intpnt{}TolPfeas".format(ptype), value)

            # Simplex primal feasibility tolerance.
            self.int.setSolverParam("basisTolX", value)

            # Mixed-integer (primal) feasibility tolerance.
            self.int.setSolverParam("mioTolFeas", value)

        # abs_dual_fsb_tol
        if self.ext.options.abs_dual_fsb_tol is not None:
            value = self.ext.options.abs_dual_fsb_tol

            # Interior-point dual feasibility tolerances.
            for ptype in ("", "Co") + (("Qo",) if mosek8 else ()):
                self.int.setSolverParam("intpnt{}TolDfeas".format(ptype), value)

            # Simplex dual feasibility (optimality) tolerance.
            self.int.setSolverParam("basisTolS", value)

        # rel_dual_fsb_tol
        if self.ext.options.rel_dual_fsb_tol is not None:
            # Simplex relative dual feasibility (optimality) tolerance.
            self.int.setSolverParam("basisRelTolS",
                self.ext.options.rel_dual_fsb_tol)

        # rel_ipm_opt_tol
        if self.ext.options.rel_ipm_opt_tol is not None:
            value = self.ext.options.rel_ipm_opt_tol

            # Interior-point primal feasibility tolerances.
            for ptype in ("", "Co") + (("Qo",) if mosek8 else ()):
                self.int.setSolverParam(
                    "intpnt{}TolRelGap".format(ptype), value)

        # abs_bnb_opt_tol
        if self.ext.options.abs_bnb_opt_tol is not None:
            self.int.setSolverParam("mioTolAbsGap",
                self.ext.options.abs_bnb_opt_tol)

        # rel_bnb_opt_tol
        if self.ext.options.rel_bnb_opt_tol is not None:
            self.int.setSolverParam("mioTolRelGap",
                self.ext.options.rel_bnb_opt_tol)

        # integrality_tol
        if self.ext.options.integrality_tol is not None:
            self.int.setSolverParam("mioTolAbsRelaxInt",
                self.ext.options.integrality_tol)

        # max_iterations
        if self.ext.options.max_iterations is not None:
            value = self.ext.options.max_iterations
            self.int.setSolverParam("biMaxIterations",     value)
            self.int.setSolverParam("intpntMaxIterations", value)
            self.int.setSolverParam("simMaxIterations",    value)

        if self.ext.options.lp_node_method is not None \
        or self.ext.options.lp_root_method is not None:
            # TODO: Give Problem an interface for checks like this.
            _islp = isinstance(
                self.ext.no.function, AffineExpression) \
                and all([isinstance(constraint, AffineConstraint)
                for constraint in self.ext.constraints.values()])

            _lpm = {
                "interior": "intpnt" if _islp else "conic",
                "psimplex": "primalSimplex",
                "dsimplex": "dualSimplex"}

        # lp_node_method
        if self.ext.options.lp_node_method is not None:
            value = self.ext.options.lp_node_method
            assert value in _lpm, "Unexpected lp_node_method value."
            self.int.setSolverParam("mioNodeOptimizer", _lpm[value])

        # lp_root_method
        if self.ext.options.lp_root_method is not None:
            value = self.ext.options.lp_root_method
            assert value in _lpm, "Unexpected lp_root_method value."
            self.int.setSolverParam("mioRootOptimizer", _lpm[value])

        # timelimit
        if self.ext.options.timelimit is not None:
            value = float(self.ext.options.timelimit)
            self.int.setSolverParam("optimizerMaxTime", value)
            self.int.setSolverParam("mioMaxTime",       value)

        # max_fsb_nodes
        if self.ext.options.max_fsb_nodes is not None:
            self.int.setSolverParam("mioMaxNumSolutions",
                self.ext.options.max_fsb_nodes)

        # hotstart
        if self.ext.options.hotstart:
            # TODO: Check if valued variables (i.e. a hotstart) are utilized by
            #       MOSEK beyond mioConstructSol, and whether it makes sense to
            #       (1) also value continuous variables and (2) reset variable
            #       values when hotstart gets disabled again (see Gurobi).
            self.int.setSolverParam("mioConstructSol", "on")
            self._import_variable_values(integralOnly=True)

        # Handle MOSEK-specific options.
        for option, value in self.ext.options.mskfsn_params.items():
            try:
                self.int.setSolverParam(option, value)
            except msk.ParameterError:
                self._handle_bad_option_value("mskfsn_params",
                    "MOSEK option '{}' does not exist.".format(option))
            except ValueError as error:
                self._handle_bad_option_value("mskfsn_params",
                    "Invalid value '{}' for MOSEK option '{}': {}"
                    .format(value, option, str(error)))

        # Handle unsupported options.
        self._handle_unsupported_option("treememory")

        # Attempt to solve the problem.
        with self._header(), self._stopwatch():
            self.int.solve()

        # Retrieve primals.
        primals = {}
        if self.ext.options.primals is not False:
            for picosVar in self.ext.variables.values():
                mosekVar = self.knownVariables[picosVar]
                try:
                    primals[picosVar] = list(mosekVar.level())
                except msk.SolutionError:
                    primals[picosVar] = None

        # Retrieve duals.
        duals = {}
        if self.ext.options.duals is not False:
            for picosCon in self.ext.constraints.values():
                # Retrieve corresponding MOSEK constraint.
                mosekCon = self.knownConstraints[picosCon]

                # Retrieve its dual.
                try:
                    mosekDual = mosekCon.dual()
                except msk.SolutionError:
                    duals[picosCon] = None
                    continue

                # Devectorize the dual. Note the change from row-major to
                # column-major order.
                # size = picosCon.size
                # picosDual = cvxopt.matrix(mosekDual, (size[1], size[0])).T
                picosDual = cvxopt.matrix(mosekDual)

                # Adjust the dual based on constraint type.
                if isinstance(picosCon, AffineConstraint) \
                or isinstance(picosCon, LMIConstraint):
                    if picosCon.is_decreasing():
                        picosDual = -picosDual
                elif isinstance(picosCon, SOCConstraint):
                    picosDual = -picosDual
                elif isinstance(picosCon, RSOCConstraint):
                    # MOSEK handles the vector [x₁; x₂; x₃] as input for a
                    # constraint of the form ‖x₃‖² ≤ 2x₁x₂ whereas PICOS handles
                    # the expressions e₁, e₂ and e₃ for a constraint of the form
                    # ‖e₁‖² ≤ e₂e₃. MOSEK's additional factor of two was
                    # neutralized on import by scaling e₂ and e₃ by sqrt(0.5)
                    # each to obtain x₁ and x₂ respectively. Scale now also the
                    # dual returned by MOSEK to make up for this.
                    scale = 0.5**0.5
                    alpha = scale * picosDual[0]
                    beta  = scale * picosDual[1]
                    z     = list(-picosDual[2:])

                    # HACK: Work around a potential documentation bug in MOSEK:
                    #       The first two vector elements of the rotated
                    #       quadratic cone dual are non-positive (allowing for a
                    #       shorter notation in the linear part of their dual
                    #       representation) even though their definition of the
                    #       (self-dual) rotated quadratic cone explicitly states
                    #       that they are non-negative (as in PICOS).
                    alpha = -alpha
                    beta  = -beta

                    picosDual = cvxopt.matrix([alpha, beta] + z)
                else:
                    assert False, \
                        "Constraint type belongs to unsupported problem type."

                # Flip sign based on objective sense.
                if (self.int.getTask().getobjsense() == objsense.minimize):
                    picosDual = -picosDual

                duals[picosCon] = picosDual

        # # Retrieve objective value.
        # try:
        #     objectiveValue = float(self.int.primalObjValue())
        # except msk.SolutionError:
        #     objectiveValue = None

        primalStatus  = self._solution_status_pic2msk(
            self.int.getPrimalSolutionStatus())
        dualStatus    = self._solution_status_pic2msk(
            self.int.getDualSolutionStatus())
        problemStatus = self._problem_status_pic2msk(self.int.getProblemStatus(
            msk.SolutionType.Default), not self.ext.is_continuous())

        # Correct two known bad solution states:
        if problemStatus == PS_INFEASIBLE and primalStatus == SS_UNKNOWN:
            primalStatus = SS_INFEASIBLE
        if problemStatus == PS_UNBOUNDED and dualStatus == SS_UNKNOWN:
            dualStatus = SS_INFEASIBLE

        return self._make_solution(
            primals, duals, primalStatus, dualStatus, problemStatus)

    def _solution_status_pic2msk(self, statusCode):
        from mosek.fusion import SolutionStatus as ss

        map = {
            ss.Undefined:       SS_EMPTY,
            ss.Unknown:         SS_UNKNOWN,
            ss.Optimal:         SS_OPTIMAL,
            ss.Feasible:        SS_FEASIBLE,
            ss.Certificate:     SS_INFEASIBLE,
            ss.IllposedCert:    SS_UNKNOWN
        }

        if self.ver <= 8:
            map.update({
                ss.NearOptimal:     SS_FEASIBLE,
                ss.NearFeasible:    SS_UNKNOWN,
                ss.NearCertificate: SS_UNKNOWN,
            })

        try:
            return map[statusCode]
        except KeyError:
            self._warn("The MOSEK Fusion solution status code {} is not known "
                "to PICOS.".format(statusCode))
            return SS_UNKNOWN

    def _problem_status_pic2msk(self, statusCode, integerProblem):
        from mosek.fusion import ProblemStatus as ps

        try:
            return {
                ps.Unknown:                     PS_UNKNOWN,
                ps.PrimalAndDualFeasible:       PS_FEASIBLE,
                ps.PrimalFeasible:
                    PS_FEASIBLE if integerProblem else PS_UNKNOWN,
                ps.DualFeasible:                PS_UNKNOWN,
                ps.PrimalInfeasible:            PS_INFEASIBLE,
                ps.DualInfeasible:              PS_UNBOUNDED,
                ps.PrimalAndDualInfeasible:     PS_INFEASIBLE,
                ps.IllPosed:                    PS_ILLPOSED,
                ps.PrimalInfeasibleOrUnbounded: PS_INF_OR_UNB
            }[statusCode]
        except KeyError:
            self._warn("The MOSEK Fusion problem status code {} is not known to"
                " PICOS.".format(statusCode))
            return PS_UNKNOWN


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`CPLEXSolver`."""

import time
from collections import namedtuple

import cvxopt

from ..apidoc import api_end, api_start
from ..compat import assert_import_exists
from ..constraints import (AffineConstraint, ConvexQuadraticConstraint,
                           RSOCConstraint, SOCConstraint)
from ..expressions import (CONTINUOUS_VARTYPES, AffineExpression,
                           BinaryVariable, IntegerVariable,
                           QuadraticExpression)
from ..modeling.footprint import Specification
from ..modeling.solution import (PS_FEASIBLE, PS_ILLPOSED, PS_INF_OR_UNB,
                                 PS_INFEASIBLE, PS_UNBOUNDED, PS_UNKNOWN,
                                 PS_UNSTABLE, SS_EMPTY, SS_FAILURE,
                                 SS_FEASIBLE, SS_INFEASIBLE, SS_OPTIMAL,
                                 SS_PREMATURE, SS_UNKNOWN)
from .solver import (ConflictingOptionsError, DependentOptionError, Solver,
                     UnsupportedOptionError)

_API_START = api_start(globals())
# -------------------------------


#: Maps CPLEX status code to PICOS status triples.
CPLEX_STATUS_CODES = {
    # primal status, dual status,   problem status
1:   (SS_OPTIMAL,    SS_OPTIMAL,    PS_FEASIBLE),    # CPX_STAT_OPTIMAL
2:   (SS_UNKNOWN,    SS_INFEASIBLE, PS_UNBOUNDED),   # CPX_STAT_UNBOUNDED
3:   (SS_INFEASIBLE, SS_UNKNOWN,    PS_INFEASIBLE),  # CPX_STAT_INFEASIBLE
4:   (SS_UNKNOWN,    SS_UNKNOWN,    PS_INF_OR_UNB),  # CPX_STAT_INForUNBD
5:   (SS_INFEASIBLE, SS_UNKNOWN,    PS_UNSTABLE),    # CPX_STAT_OPTIMAL_INFEAS
6:   (SS_UNKNOWN,    SS_UNKNOWN,    PS_UNSTABLE),    # CPX_STAT_NUM_BEST
    # 7—9 are not defined.
10:  (SS_PREMATURE,  SS_PREMATURE,  PS_UNKNOWN),     # CPX_STAT_ABORT_IT_LIM
11:  (SS_PREMATURE,  SS_PREMATURE,  PS_UNKNOWN),     # CPX_STAT_ABORT_TIME_LIM
12:  (SS_PREMATURE,  SS_PREMATURE,  PS_UNKNOWN),     # CPX_STAT_ABORT_OBJ_LIM
13:  (SS_PREMATURE,  SS_PREMATURE,  PS_UNKNOWN),     # CPX_STAT_ABORT_USER
    # 14—19 seem irrelevant (CPX_STAT_*_RELAXED_*).
20:  (SS_UNKNOWN,    SS_UNKNOWN,    PS_ILLPOSED),    # …_OPTIMAL_FACE_UNBOUNDED
21:  (SS_PREMATURE,  SS_PREMATURE,  PS_UNKNOWN),     # …_ABORT_PRIM_OBJ_LIM
22:  (SS_PREMATURE,  SS_PREMATURE,  PS_UNKNOWN),     # …_ABORT_DUAL_OBJ_LIM
23:  (SS_FEASIBLE,   SS_FEASIBLE,   PS_FEASIBLE),    # CPX_STAT_FEASIBLE
    # 24 irrelevant (CPX_STAT_FIRSTORDER).
25:  (SS_PREMATURE,  SS_PREMATURE,  PS_UNKNOWN),     # …_ABORT_DETTIME_LIM
    # 26—29 are not defined.
    # 30—39 seem irrelevant (CPX_STAT_CONFLICT_*).
    # 40—100 are not defined.
101: (SS_OPTIMAL,    SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_OPTIMAL
102: (SS_OPTIMAL,    SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_OPTIMAL_TOL
103: (SS_INFEASIBLE, SS_EMPTY,      PS_INFEASIBLE),  # CPXMIP_INFEASIBLE
104: (SS_PREMATURE,  SS_EMPTY,      PS_UNKNOWN),     # CPXMIP_SOL_LIM          ?
105: (SS_FEASIBLE,   SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_NODE_LIM_FEAS
106: (SS_PREMATURE,  SS_EMPTY,      PS_UNKNOWN),     # CPXMIP_NODE_LIM_INFEAS
107: (SS_FEASIBLE,   SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_TIME_LIM_FEAS
108: (SS_PREMATURE,  SS_EMPTY,      PS_UNKNOWN),     # CPXMIP_TIME_LIM_INFEAS
109: (SS_FEASIBLE,   SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_FAIL_FEAS
110: (SS_FAILURE,    SS_EMPTY,      PS_UNKNOWN),     # CPXMIP_FAIL_INFEAS
111: (SS_FEASIBLE,   SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_MEM_LIM_FEAS
112: (SS_PREMATURE,  SS_EMPTY,      PS_UNKNOWN),     # CPXMIP_MEM_LIM_INFEAS
113: (SS_FEASIBLE,   SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_ABORT_FEAS
114: (SS_PREMATURE,  SS_EMPTY,      PS_UNKNOWN),     # CPXMIP_ABORT_INFEAS
115: (SS_INFEASIBLE, SS_EMPTY,      PS_UNSTABLE),    # CPXMIP_OPTIMAL_INFEAS
116: (SS_FEASIBLE,   SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_FAIL_FEAS_NO_TREE
117: (SS_FAILURE,    SS_EMPTY,      PS_UNKNOWN),     # …_FAIL_INFEAS_NO_TREE
118: (SS_UNKNOWN,    SS_EMPTY,      PS_UNBOUNDED),   # CPXMIP_UNBOUNDED
119: (SS_UNKNOWN,    SS_EMPTY,      PS_INF_OR_UNB),  # CPXMIP_INForUNBD
    # 120—126 seem irrelevant (CPXMIP_*_RELAXED_*).
127: (SS_FEASIBLE,   SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_FEASIBLE
128: (SS_OPTIMAL,    SS_EMPTY,      PS_FEASIBLE),    # …_POPULATESOL_LIM       ?
129: (SS_OPTIMAL,    SS_EMPTY,      PS_FEASIBLE),    # …_OPTIMAL_POPULATED     ?
130: (SS_OPTIMAL,    SS_EMPTY,      PS_FEASIBLE),    # …_OPTIMAL_POPULATED_TOL ?
131: (SS_FEASIBLE,   SS_EMPTY,      PS_FEASIBLE),    # CPXMIP_DETTIME_LIM_FEAS
132: (SS_PREMATURE,  SS_EMPTY,      PS_UNKNOWN),     # CPXMIP_DETTIME_LIM_INFEAS
}


class CPLEXSolver(Solver):
    """Interface to the CPLEX solver via its official Python interface.

    .. note ::
        Names are used instead of indices for identifying both variables and
        constraints since indices can change if the CPLEX instance is modified.
    """

    # TODO: Allow nonconvex quadratic constraints in the integer case?
    SUPPORTED = Specification.compile(
        directions=Specification.ALL,
        objectives=[
            AffineExpression,
            QuadraticExpression],
        variables=Specification.ALL,
        constraints=[
            AffineConstraint,
            SOCConstraint,
            RSOCConstraint,
            ConvexQuadraticConstraint])

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.solver.Solver.supports`."""
        if not Solver.supports(footprint):
            return False

        if footprint.nonconvex_quadratic_objective and footprint.continuous:
            return False

        return footprint << cls.SUPPORTED

    @classmethod
    def default_penalty(cls):
        """Implement :meth:`~.solver.Solver.default_penalty`."""
        return 0.0  # Commercial solver.

    @classmethod
    def test_availability(cls):
        """Implement :meth:`~.solver.Solver.test_availability`."""
        assert_import_exists("cplex")

    @classmethod
    def names(cls):
        """Implement :meth:`~.solver.Solver.names`."""
        return "cplex", "CPLEX", "IBM ILOG CPLEX Optimization Studio"

    CplexSOCC = namedtuple("CplexSOCC",
        ("LHSVars", "RHSVar", "LHSCons", "RHSCon", "quadCon"))

    CplexRSOCC = namedtuple("CplexRSOCC",
        ("LHSVars", "RHSVars", "LHSCons", "RHSCons", "quadCon"))

    def __init__(self, problem):
        """Initialize a CPLEX solver interface.

        :param ~picos.Problem problem: The problem to be solved.
        """
        super(CPLEXSolver, self).__init__(problem)

        self._cplexVarName = dict()
        """Maps PICOS variable indices to CPLEX variable names."""

        self._cplexLinConNames = dict()
        """Maps a PICOS (multidimensional) linear constraint to a collection of
        CPLEX (scalar) linear constraint names."""

        self._cplexQuadConName = dict()
        """Maps a PICOS quadratic or conic quadratic constraint to a CPLEX
        quadratic constraint name."""

        self._cplexSOCC = dict()
        """Maps a PICOS second order cone constraint to its CPLEX representation
        involving auxiliary variables and constraints."""

        self._cplexRSOCC = dict()
        """Maps a PICOS rotated second order cone constraint to its CPLEX
        representation involving auxiliary variables and constraints."""

        self.nextConstraintID = 0
        """Used to create unique names for constraints."""

    def __del__(self):
        if self.int is not None:
            self.int.end()

    def reset_problem(self):
        """Implement :meth:`~.solver.Solver.reset_problem`."""
        if self.int is not None:
            self.int.end()
        self.int = None
        self._cplexVarName.clear()
        self._cplexLinConNames.clear()
        self._cplexQuadConName.clear()
        self._cplexSOCC.clear()
        self._cplexRSOCC.clear()

    def _get_unique_constraint_id(self):
        ID = self.nextConstraintID
        self.nextConstraintID += 1
        return ID

    def _make_cplex_var_names(self, picosVar, localIndex=None):
        """Make CPLEX variable names.

        Converts a PICOS variable to a list of CPLEX variable names, each
        corresponding to one scalar variable contained in the PICOS variable.
        If localIndex is given, then only the name of the CPLEX variable
        representing the scalar variable with that offset is returned.
        The name format is "picosName[localIndex]".
        """
        # TODO: This function appears in multiple solvers, move it to the Solver
        #       base class as "_make_scalar_var_names".
        if localIndex is not None:
            return "{}[{}]".format(picosVar.name, localIndex)
        else:
            return [
                self._make_cplex_var_names(picosVar, localIndex)
                for localIndex in range(picosVar.dim)]

    def _import_variable(self, picosVar):
        import cplex

        dim = picosVar.dim

        # Create names.
        names = self._make_cplex_var_names(picosVar)

        # Retrieve types.
        if isinstance(picosVar, CONTINUOUS_VARTYPES):
            types = dim * self.int.variables.type.continuous
        elif isinstance(picosVar, IntegerVariable):
            types = dim * self.int.variables.type.integer
        elif isinstance(picosVar, BinaryVariable):
            types = dim * self.int.variables.type.binary
        else:
            assert False, "Unexpected variable type."

        # Retrieve bounds.
        lowerBounds = [-cplex.infinity]*dim
        upperBounds = [cplex.infinity]*dim
        lower, upper = picosVar.bound_dicts
        for i, b in lower.items():
            lowerBounds[i] = b
        for i, b in upper.items():
            upperBounds[i] = b

        # Import variable.
        # Note that CPLEX allows importing the objective function coefficients
        # for the new variables here, but that is done later to streamline
        # updates to the objective.
        self.int.variables.add(
            lb=lowerBounds, ub=upperBounds, types=types, names=names)

        # Map PICOS indices to CPLEX names.
        for localIndex in range(dim):
            self._cplexVarName[picosVar.id_at(localIndex)] = names[localIndex]

        if self._debug():
            cplexVar = {"names": names, "types": types,
                "lowerBounds": lowerBounds, "upperBounds": upperBounds}
            self._debug(
                "Variable imported: {} → {}".format(picosVar, cplexVar))

    def _remove_variable(self, picosVar):
        cplexVarNames = [self._cplexVarName.pop(picosVar.id_at(localIndex))
            for localIndex in range(picosVar.dim)]
        self.int.variables.delete(cplexVarNames)

    def _affinexp_pic2cpl(self, picosExpression):
        import cplex
        for names, coefficients, constant in picosExpression.sparse_rows(
                None, indexFunction=self._make_cplex_var_names):
            yield cplex.SparsePair(ind=names, val=coefficients), constant

    def _scalar_affinexp_pic2cpl(self, picosExpression):
        assert len(picosExpression) == 1
        return next(self._affinexp_pic2cpl(picosExpression))

    def _quadexp_pic2cpl(self, picosExpression):
        """Transform a quadratic expression from PICOS to CPLEX.

        :returns: :class:`SparseTriple <cplex.SparseTriple>` mapping a pair of
            CPLEX variable names to scalar constants.
        """
        import cplex

        assert isinstance(picosExpression, QuadraticExpression)

        cplexI, cplexJ, cplexV = [], [], []
        for (picosVar1, picosVar2), picosCoefficients \
                in picosExpression.quadratic_forms.items():
            for sparseIndex in range(len(picosCoefficients)):
                localVar1Index = picosCoefficients.I[sparseIndex]
                localVar2Index = picosCoefficients.J[sparseIndex]
                localCoefficient = picosCoefficients.V[sparseIndex]
                cplexI.append(self._cplexVarName[picosVar1.id + localVar1Index])
                cplexJ.append(self._cplexVarName[picosVar2.id + localVar2Index])
                cplexV.append(localCoefficient)

        return cplex.SparseTriple(ind1=cplexI, ind2=cplexJ, val=cplexV)

    def _import_linear_constraint(self, picosConstraint):
        import cplex

        assert isinstance(picosConstraint, AffineConstraint)

        length = len(picosConstraint)

        # Retrieve left hand side and right hand side expressions.
        cplexLHS, cplexRHS = [], []
        for names, coefficients, constant in picosConstraint.sparse_Ab_rows(
                None, indexFunction=self._make_cplex_var_names):
            cplexLHS.append(cplex.SparsePair(ind=names, val=coefficients))
            cplexRHS.append(constant)

        # Retrieve senses.
        if picosConstraint.is_increasing():
            senses = length * "L"
        elif picosConstraint.is_decreasing():
            senses = length * "G"
        elif picosConstraint.is_equality():
            senses = length * "E"
        else:
            assert False, "Unexpected constraint relation."

        # Give unique names that are used to identify the constraint. This is
        # necessary as constraint indices can change if the problem is modified.
        conID = self._get_unique_constraint_id()
        names = ["{}:{}".format(conID, localConstraintIndex)
            for localConstraintIndex in range(length)]

        if self._debug():
            cplexConstraint = {"lin_expr": cplexLHS, "senses": senses,
                "rhs": cplexRHS, "names": names}
            self._debug(
                "Linear constraint imported: {} → {}".format(
                    picosConstraint, cplexConstraint))

        # Import the constraint.
        self.int.linear_constraints.add(
            lin_expr=cplexLHS, senses=senses, rhs=cplexRHS, names=names)

        return names

    def _import_quad_constraint(self, picosConstraint):
        assert isinstance(picosConstraint, ConvexQuadraticConstraint)

        # Retrieve
        # - CPLEX' linear term, which is the linear term of the affine
        #   expression in the PICOS constraint, and
        # - CPLEX' right hand side, which is the negated constant term of the
        #   affine expression in the PICOS constraint.
        cplexLinear, cplexRHS = \
            self._scalar_affinexp_pic2cpl(picosConstraint.le0.aff)
        cplexRHS = -cplexRHS

        # Retrieve CPLEX' quadratic term.
        cplexQuad = self._quadexp_pic2cpl(picosConstraint.le0)

        # Give a unique name that is used to identify the constraint. This is
        # necessary as constraint indices can change if the problem is modified.
        name = "{}:{}".format(self._get_unique_constraint_id(), 0)

        if self._debug():
            cplexConstraint = {"lin_expr": cplexLinear, "quad_expr": cplexQuad,
                "rhs": cplexRHS, "name": name}
            self._debug(
                "Quadratic constraint imported: {} → {}".format(
                    picosConstraint, cplexConstraint))

        # Import the constraint.
        self.int.quadratic_constraints.add(lin_expr=cplexLinear,
            quad_expr=cplexQuad, sense="L", rhs=cplexRHS, name=name)

        return name

    # TODO: Handle SOC → Quadratic via a reformulation.
    def _import_socone_constraint(self, picosConstraint):
        import cplex

        assert isinstance(picosConstraint, SOCConstraint)

        picosLHS = picosConstraint.ne
        picosRHS = picosConstraint.ub
        picosLHSLen = len(picosLHS)

        # Make identifying names for the auxiliary variables and constraints.
        conID = self._get_unique_constraint_id()
        cplexLHSVars = ["{}:V{}".format(conID, i) for i in range(picosLHSLen)]
        cplexRHSVar  = "{}:V{}".format(conID, picosLHSLen)
        cplexLHSCons = ["{}:C{}".format(conID, i) for i in range(picosLHSLen)]
        cplexRHSCon  = "{}:C{}".format(conID, picosLHSLen)
        cplexQuadCon = "{}:C{}".format(conID, picosLHSLen + 1)

        # Add auxiliary variables: One for every dimension of the left hand side
        # of the PICOS constraint and one for its right hand side.
        self.int.variables.add(
            names=cplexLHSVars, lb=[-cplex.infinity] * picosLHSLen,
            ub=[+cplex.infinity] * picosLHSLen,
            types=self.int.variables.type.continuous * picosLHSLen)
        self.int.variables.add(
            names=[cplexRHSVar], lb=[0.0], ub=[+cplex.infinity],
            types=self.int.variables.type.continuous)

        # Add constraints that identify the left hand side CPLEX auxiliary
        # variables with their slice of the PICOS left hand side expression.
        cplexLHSConsLHSs = []
        cplexLHSConsRHSs = []
        for localConIndex, (localLinExp, localConstant) in \
                enumerate(self._affinexp_pic2cpl(picosLHS)):
            localConstant = -localConstant
            localLinExp.ind.append(cplexLHSVars[localConIndex])
            localLinExp.val.append(-1.0)
            cplexLHSConsLHSs.append(localLinExp)
            cplexLHSConsRHSs.append(localConstant)
        self.int.linear_constraints.add(
            names=cplexLHSCons, lin_expr=cplexLHSConsLHSs,
            senses="E" * picosLHSLen, rhs=cplexLHSConsRHSs)

        # Add a constraint that identifies the right hand side CPLEX auxiliary
        # variable with the PICOS right hand side scalar expression.
        cplexRHSConLHS, cplexRHSConRHS = \
            self._scalar_affinexp_pic2cpl(-picosRHS)
        cplexRHSConRHS = -cplexRHSConRHS
        cplexRHSConLHS.ind.append(cplexRHSVar)
        cplexRHSConLHS.val.append(1.0)
        self.int.linear_constraints.add(
            names=[cplexRHSCon], lin_expr=[cplexRHSConLHS],
            senses="E", rhs=[cplexRHSConRHS])

        # Add a quadratic constraint over the auxiliary variables that
        # represents the PICOS second order cone constraint itself.
        quadIndices = [cplexRHSVar] + list(cplexLHSVars)
        quadExpr = cplex.SparseTriple(
            ind1=quadIndices, ind2=quadIndices, val=[-1.0] + [1.0]*picosLHSLen)
        self.int.quadratic_constraints.add(
            name=cplexQuadCon, quad_expr=quadExpr, sense="L", rhs=0.0)

        cplexMetaCon = self.CplexSOCC(LHSVars=cplexLHSVars, RHSVar=cplexRHSVar,
            LHSCons=cplexLHSCons, RHSCon=cplexRHSCon, quadCon=cplexQuadCon)

        if self._debug():
            cplexCons = {
                "LHSs of LHS auxiliary equalities": cplexLHSConsLHSs,
                "RHSs of LHS auxiliary equalities": cplexLHSConsRHSs,
                "LHS of RHS auxiliary equality": cplexRHSConLHS,
                "RHS of RHS auxiliary equality": cplexRHSConRHS,
                "Non-positive quadratic term": quadExpr}
            self._debug(
                "SOcone constraint imported: {} → {}, {}".format(
                    picosConstraint, cplexMetaCon, cplexCons))

        return cplexMetaCon

    # TODO: Handle RSOC → Quadratic via a reformulation.
    def _import_rscone_constraint(self, picosConstraint):
        import cplex

        assert isinstance(picosConstraint, RSOCConstraint)

        picosLHS = picosConstraint.ne
        picosRHS1 = picosConstraint.ub1
        picosRHS2 = picosConstraint.ub2
        picosLHSLen = len(picosLHS)

        # Make identifying names for the auxiliary variables and constraints.
        conID = self._get_unique_constraint_id()
        cplexLHSVars = ["{}:V{}".format(conID, i) for i in range(picosLHSLen)]
        cplexRHSVars = ["{}:V{}".format(conID, picosLHSLen + i) for i in (0, 1)]
        cplexLHSCons = ["{}:C{}".format(conID, i) for i in range(picosLHSLen)]
        cplexRHSCons = ["{}:C{}".format(conID, picosLHSLen + i) for i in (0, 1)]
        cplexQuadCon = "{}:C{}".format(conID, picosLHSLen + 2)

        # Add auxiliary variables: One for every dimension of the left hand side
        # of the PICOS constraint and two for its right hand side.
        self.int.variables.add(
            names=cplexLHSVars, lb=[-cplex.infinity] * picosLHSLen,
            ub=[+cplex.infinity] * picosLHSLen,
            types=self.int.variables.type.continuous * picosLHSLen)
        self.int.variables.add(
            names=cplexRHSVars, lb=[0.0, 0.0], ub=[+cplex.infinity] * 2,
            types=self.int.variables.type.continuous * 2)

        # Add constraints that identify the left hand side CPLEX auxiliary
        # variables with their slice of the PICOS left hand side expression.
        cplexLHSConsLHSs = []
        cplexLHSConsRHSs = []
        for localConIndex, (localLinExp, localConstant) in \
                enumerate(self._affinexp_pic2cpl(picosLHS)):
            localLinExp.ind.append(cplexLHSVars[localConIndex])
            localLinExp.val.append(-1.0)
            localConstant = -localConstant
            cplexLHSConsLHSs.append(localLinExp)
            cplexLHSConsRHSs.append(localConstant)
        self.int.linear_constraints.add(
            names=cplexLHSCons, lin_expr=cplexLHSConsLHSs,
            senses="E" * picosLHSLen, rhs=cplexLHSConsRHSs)

        # Add two constraints that identify the right hand side CPLEX auxiliary
        # variables with the PICOS right hand side scalar expressions.
        cplexRHSConsLHSs = []
        cplexRHSConsRHSs = []
        for picosRHS, cplexRHSVar in zip((picosRHS1, picosRHS2), cplexRHSVars):
            linExp, constant = self._scalar_affinexp_pic2cpl(-picosRHS)
            linExp.ind.append(cplexRHSVar)
            linExp.val.append(1.0)
            constant = -constant
            cplexRHSConsLHSs.append(linExp)
            cplexRHSConsRHSs.append(constant)
        self.int.linear_constraints.add(
            names=cplexRHSCons, lin_expr=cplexRHSConsLHSs,
            senses="E" * 2, rhs=cplexRHSConsRHSs)

        # Add a quadratic constraint over the auxiliary variables that
        # represents the PICOS rotated second order cone constraint itself.
        quadExpr = cplex.SparseTriple(
            ind1=[cplexRHSVars[0]] + list(cplexLHSVars),
            ind2=[cplexRHSVars[1]] + list(cplexLHSVars),
            val=[-1.0] + [1.0] * picosLHSLen)
        self.int.quadratic_constraints.add(
            name=cplexQuadCon, quad_expr=quadExpr, sense="L", rhs=0.0)

        cplexMetaCon = self.CplexRSOCC(
            LHSVars=cplexLHSVars, RHSVars=cplexRHSVars, LHSCons=cplexLHSCons,
            RHSCons=cplexRHSCons, quadCon=cplexQuadCon)

        if self._debug():
            cplexCons = {
                "LHSs of LHS auxiliary equalities": cplexLHSConsLHSs,
                "RHSs of LHS auxiliary equalities": cplexLHSConsRHSs,
                "LHSs of RHS auxiliary equalities": cplexRHSConsLHSs,
                "RHSs of RHS auxiliary equalities": cplexRHSConsRHSs,
                "Non-positive quadratic term": quadExpr}
            self._debug(
                "RScone constraint imported: {} → {}, {}".format(
                    picosConstraint, cplexMetaCon, cplexCons))

        return cplexMetaCon

    def _import_constraint(self, picosConstraint):
        # Import constraint based on type and keep track of the corresponding
        # CPLEX constraint and auxiliary variable names.
        if isinstance(picosConstraint, AffineConstraint):
            self._cplexLinConNames[picosConstraint] = \
                self._import_linear_constraint(picosConstraint)
        elif isinstance(picosConstraint, ConvexQuadraticConstraint):
            self._cplexQuadConName[picosConstraint] = \
                self._import_quad_constraint(picosConstraint)
        elif isinstance(picosConstraint, SOCConstraint):
            self._cplexSOCC[picosConstraint] = \
                self._import_socone_constraint(picosConstraint)
        elif isinstance(picosConstraint, RSOCConstraint):
            self._cplexRSOCC[picosConstraint] = \
                self._import_rscone_constraint(picosConstraint)
        else:
            assert False, "Constraint type belongs to unsupported problem type."

    def _remove_constraint(self, picosConstraint):
        if isinstance(picosConstraint, AffineConstraint):
            self.int.linear_constraints.delete(
                self._cplexLinConNames.pop(picosConstraint))
        elif isinstance(picosConstraint, ConvexQuadraticConstraint):
            self.int.quadratic_constraints.delete(
                self._cplexQuadConName.pop(picosConstraint))
        elif isinstance(picosConstraint, SOCConstraint):
            c = self._cplexSOCC.pop(picosConstraint)
            self.int.linear_constraints.delete(c.cplexLHSCons + [c.cplexRHSCon])
            self.int.quadratic_constraints.delete(c.cplexQuadCon)
            self.int.variables.delete(c.cplexLHSVars + [c.cplexRHSVar])
        elif isinstance(picosConstraint, RSOCConstraint):
            c = self._cplexRSOCC.pop(picosConstraint)
            self.int.linear_constraints.delete(c.cplexLHSCons + c.cplexRHSCons)
            self.int.quadratic_constraints.delete(c.cplexQuadCon)
            self.int.variables.delete(c.cplexLHSVars + c.cplexRHSVars)
        else:
            assert False, "Constraint type belongs to unsupported problem type."

    def _import_affine_objective(self, picosExpression):
        assert isinstance(picosExpression, AffineExpression)

        cplexExpression = []
        for picosVar, picosCoefficient in picosExpression._coefs.items():
            assert picosCoefficient.size[0] == 1

            for localIndex in range(picosVar.dim):
                cplexCoefficient = picosCoefficient[localIndex]
                if not cplexCoefficient:
                    continue
                picosIndex = picosVar.id + localIndex
                cplexName = self._cplexVarName[picosIndex]
                cplexExpression.append((cplexName, cplexCoefficient))

        if cplexExpression:
            self.int.objective.set_linear(cplexExpression)

            if self._debug():
                self._debug("Affine (part of) objective imported: {} → {}"
                    .format(picosExpression, cplexExpression))

    def _reset_affine_objective(self):
        linear = self.int.objective.get_linear()
        if any(linear):
            self.int.objective.set_linear(
                [(cplexVarIndex, 0.0) for cplexVarIndex, coefficient
                in enumerate(linear) if coefficient])

    def _import_quadratic_objective(self, picosExpression):
        assert isinstance(picosExpression, QuadraticExpression)

        # Import affine part of objective function.
        self._import_affine_objective(picosExpression.aff)

        # Import quadratic part of objective function.
        cplexQuadExpression = self._quadexp_pic2cpl(picosExpression)
        cplexQuadCoefficients = zip(
            cplexQuadExpression.ind1, cplexQuadExpression.ind2,
            [2.0 * coefficient for coefficient in cplexQuadExpression.val])
        self.int.objective.set_quadratic_coefficients(cplexQuadCoefficients)

    def _reset_quadratic_objective(self):
        quadratics = self.int.objective.get_quadratic()
        if quadratics:
            self.int.objective.set_quadratic(
                [(sparsePair.ind, [0]*len(sparsePair.ind))
                for sparsePair in quadratics])

    def _import_objective(self):
        picosSense, picosObjective = self.ext.no

        # Import objective sense.
        if picosSense == "min":
            cplexSense = self.int.objective.sense.minimize
        else:
            assert picosSense == "max"
            cplexSense = self.int.objective.sense.maximize
        self.int.objective.set_sense(cplexSense)

        # Import objective function.
        if isinstance(picosObjective, AffineExpression):
            self._import_affine_objective(picosObjective)
        else:
            assert isinstance(picosObjective, QuadraticExpression)
            self._import_quadratic_objective(picosObjective)

    def _reset_objective(self):
        self._reset_affine_objective()
        self._reset_quadratic_objective()

    def _import_problem(self):
        import cplex

        # Create a problem instance.
        self.int = cplex.Cplex()

        # Import variables.
        for variable in self.ext.variables.values():
            self._import_variable(variable)

        # Import constraints.
        for constraint in self.ext.constraints.values():
            self._import_constraint(constraint)

        # Set objective.
        self._import_objective()

    def _update_problem(self):
        for oldConstraint in self._removed_constraints():
            self._remove_constraint(oldConstraint)

        for oldVariable in self._removed_variables():
            self._remove_variable(oldVariable)

        for newVariable in self._new_variables():
            self._import_variable(newVariable)

        for newConstraint in self._new_constraints():
            self._import_constraint(newConstraint)

        if self._objective_has_changed():
            self._reset_objective()
            self._import_objective()

    def _solve(self):
        import cplex

        # Reset options.
        self.int.parameters.reset()

        o = self.ext.options
        p = self.int.parameters

        continuous = self.ext.is_continuous()

        # verbosity
        verbosity = self.verbosity()
        if verbosity <= 0:
            # Note that this behaviour disables warning even with a verbosity of
            # zero but this is still better than having verbose output for every
            # option that is set.
            self.int.set_results_stream(None)
        else:
            p.barrier.display.set(min(2, verbosity))
            p.conflict.display.set(min(2, verbosity))
            p.mip.display.set(min(5, verbosity))
            p.sifting.display.set(min(2, verbosity))
            p.simplex.display.set(min(2, verbosity))
            p.tune.display.set(min(3, verbosity))
        self.int.set_error_stream(None)  # Already handled as exceptions.

        # abs_prim_fsb_tol
        if o.abs_prim_fsb_tol is not None:
            p.simplex.tolerances.feasibility.set(o.abs_prim_fsb_tol)

        # abs_dual_fsb_tol
        if o.abs_dual_fsb_tol is not None:
            p.simplex.tolerances.optimality.set(o.abs_dual_fsb_tol)

        # rel_prim_fsb_tol, rel_dual_fsb_tol, rel_ipm_opt_tol
        convergenceTols = [tol for tol in (o.rel_prim_fsb_tol,
            o.rel_dual_fsb_tol, o.rel_ipm_opt_tol) if tol is not None]
        if convergenceTols:
            convergenceTol = min(convergenceTols)
            p.barrier.convergetol.set(convergenceTol)
            p.barrier.qcpconvergetol.set(convergenceTol)

        # abs_bnb_opt_tol
        if o.abs_bnb_opt_tol is not None:
            p.mip.tolerances.absmipgap.set(o.abs_bnb_opt_tol)

        # rel_bnb_opt_tol
        if o.rel_bnb_opt_tol is not None:
            p.mip.tolerances.mipgap.set(o.rel_bnb_opt_tol)

        # integrality_tol
        if o.integrality_tol is not None:
            p.mip.tolerances.integrality.set(o.integrality_tol)

        # markowitz_tol
        if o.markowitz_tol is not None:
            p.simplex.tolerances.markowitz.set(o.markowitz_tol)

        # max_iterations
        if o.max_iterations is not None:
            maxit = o.max_iterations
            p.barrier.limits.iteration.set(maxit)
            p.simplex.limits.iterations.set(maxit)

        _lpm = {"interior": 4, "psimplex": 1, "dsimplex": 2}

        # lp_node_method
        if o.lp_node_method is not None:
            assert o.lp_node_method in _lpm, "Unexpected lp_node_method value."
            p.mip.strategy.subalgorithm.set(_lpm[o.lp_node_method])

        # lp_root_method
        if o.lp_root_method is not None:
            assert o.lp_root_method in _lpm, "Unexpected lp_root_method value."
            p.lpmethod.set(_lpm[o.lp_root_method])

        # timelimit
        if o.timelimit is not None:
            p.timelimit.set(o.timelimit)

        # treememory
        if o.treememory is not None:
            p.mip.limits.treememory.set(o.treememory)

        # Handle option conflict between "max_fsb_nodes" and "pool_size".
        if o.max_fsb_nodes is not None \
        and o.pool_size is not None:
            raise ConflictingOptionsError("The options 'max_fsb_nodes' and "
                "'pool_size' cannot be used in conjunction.")

        # max_fsb_nodes
        if o.max_fsb_nodes is not None:
            p.mip.limits.solutions.set(o.max_fsb_nodes)

        # pool_size
        if o.pool_size is not None:
            if continuous:
                raise UnsupportedOptionError("The option 'pool_size' can only "
                    "be used with mixed integer problems.")
            maxNumSolutions = max(1, int(o.pool_size))
            p.mip.limits.populate.set(maxNumSolutions)
        else:
            maxNumSolutions = 1

        # pool_relgap
        if o.pool_rel_gap is not None:
            if o.pool_size is None:
                raise DependentOptionError("The option 'pool_rel_gap' requires "
                    "the option 'pool_size'.")
            p.mip.pool.relgap.set(o.pool_rel_gap)

        # pool_abs_gap
        if o.pool_abs_gap is not None:
            if o.pool_size is None:
                raise DependentOptionError("The option 'pool_abs_gap' requires "
                    "the option 'pool_size'.")
            p.mip.pool.absgap.set(o.pool_abs_gap)

        # hotstart
        if o.hotstart:
            names, values = [], []
            for picosVar in self.ext.variables.values():
                if picosVar.valued:
                    for localIndex in range(picosVar.dim):
                        name = self._cplexVarName[picosVar.id_at(localIndex)]
                        names.append(name)
                        values.append(picosVar.internal_value[localIndex])
            if names:
                self.int.MIP_starts.add(
                    cplex.SparsePair(ind=names, val=values),
                    self.int.MIP_starts.effort_level.repair)

        # Handle CPLEX-specific options.
        for option, value in o.cplex_params.items():
            try:
                parameter = eval("self.int.parameters." + option)
                parameter.set(value)
            except AttributeError:
                self._handle_bad_option_value("cplex_params",
                    "CPLEX option '{}' does not exist.".format(option))
            except cplex.exceptions.errors.CplexError:
                self._handle_bad_option_value("cplex_params",
                    "Invalid value '{}' for CPLEX option '{}'."
                    .format(value, option))

        # Handle options "cplex_upr_bnd_limit", "cplex_lwr_bnd_limit" and
        # "cplex_bnd_monitor" via a CPLEX callback handler.
        callback = None
        if o.cplex_upr_bnd_limit or o.cplex_lwr_bnd_limit \
        or o.cplex_bnd_monitor:
            from cplex.callbacks import MIPInfoCallback

            class PicosInfoCallback(MIPInfoCallback):
                def __call__(self):
                    v1 = self.get_incumbent_objective_value()
                    v2 = self.get_best_objective_value()
                    ub = max(v1, v2)
                    lb = min(v1, v2)
                    if self.bounds is not None:
                        elapsedTime = time.time() - self.startTime
                        self.bounds.append((elapsedTime, lb, ub))
                    if self.lbound is not None and lb >= self.lbound:
                        self.printer("The specified lower bound was reached, "
                            "so PICOS will ask CPLEX to stop the search.")
                        self.abort()
                    if self.ubound is not None and ub <= self.ubound:
                        self.printer("The specified upper bound was reached, "
                            "so PICOS will ask CPLEX to stop the search.")
                        self.abort()

            # Register the callback handler with CPLEX.
            callback = self.int.register_callback(PicosInfoCallback)

            # Pass parameters to the callback handler. Note that
            # callback.startTime will be set just before optimization begins.
            callback.printer = self._verbose
            callback.ubound = o.cplex_upr_bnd_limit
            callback.lbound = o.cplex_lwr_bnd_limit
            callback.bounds = [] if o.cplex_bnd_monitor else None

        # Inform CPLEX about the problem type.
        # This seems necessary, as otherwise LP can get solved as MIP, producing
        # misleading status output (e.g. "not integer feasible").
        conTypes   = set(c.__class__ for c in self.ext.constraints.values())
        quadObj    = isinstance(self.ext.no.function, QuadraticExpression)
        continuous = continuous
        cplexTypes = self.int.problem_type

        if not conTypes and quadObj:
            cplexType = cplexTypes.QP if continuous else cplexTypes.MIQP
        elif conTypes.issubset(set([AffineConstraint])):
            cplexType = cplexTypes.LP if continuous else cplexTypes.MILP
        elif conTypes.issubset(set([AffineConstraint, SOCConstraint,
                RSOCConstraint, ConvexQuadraticConstraint])):
            cplexType = cplexTypes.QCP if continuous else cplexTypes.MIQCP
        else:
            cplexType = None
            self._verbose(
                "PICOS is not sure how to inform CPLEX that the problem type is"
                "'{}', and will trust CPLEX to detect this correctly."
                .format(self.int.type))

        if cplexType is not None:
            self.int.set_problem_type(cplexType)

        # Attempt to solve the problem.
        if callback:
            callback.startTime = time.time()
        with self._header(), self._stopwatch():
            try:
                if maxNumSolutions > 1:
                    self.int.populate_solution_pool()
                    numSolutions = self.int.solution.pool.get_num()
                else:
                    self.int.solve()
                    numSolutions = 1
            except cplex.exceptions.errors.CplexSolverError as error:
                if error.args[2] == 5002:
                    self._handle_continuous_nonconvex_error(error)
                else:
                    raise

        solutions = []
        for solutionNum in range(numSolutions):
            # Retrieve primals.
            primals = {}
            if o.primals is not False:
                for picosVar in self.ext.variables.values():
                    try:
                        cplexNames = []

                        for localIndex in range(picosVar.dim):
                            picosIndex = picosVar.id + localIndex
                            cplexNames.append(self._cplexVarName[picosIndex])
                        if maxNumSolutions > 1:
                            value = self.int.solution.pool.get_values(
                                solutionNum, cplexNames)
                        else:
                            value = self.int.solution.get_values(cplexNames)
                        primals[picosVar] = value
                    except cplex.exceptions.errors.CplexSolverError:
                        primals[picosVar] = None

            # Retrieve duals.
            duals = {}
            if o.duals is not False and continuous:
                assert maxNumSolutions == 1

                for picosCon in self.ext.constraints.values():
                    try:
                        if isinstance(picosCon, AffineConstraint):
                            cplexCons = self._cplexLinConNames[picosCon]
                            values = self.int.solution.get_dual_values(
                                cplexCons)
                            picosDual = cvxopt.matrix(values, picosCon.size)

                            # Flip sign based on constraint relation.
                            if picosCon.is_decreasing():
                                picosDual = -picosDual
                        elif isinstance(picosCon, SOCConstraint):
                            cplexMetaCon = self._cplexSOCC[picosCon]
                            lb = self.int.solution.get_dual_values(
                                cplexMetaCon.RHSCon)
                            z = self.int.solution.get_dual_values(
                                list(cplexMetaCon.LHSCons))
                            picosDual = -cvxopt.matrix([-lb] + z)
                        elif isinstance(picosCon, RSOCConstraint):
                            cplexMetaCon = self._cplexRSOCC[picosCon]
                            ab = [-x for x in self.int.solution.get_dual_values(
                                list(cplexMetaCon.RHSCons))]
                            z = self.int.solution.get_dual_values(
                                list(cplexMetaCon.LHSCons))
                            picosDual = -cvxopt.matrix(ab + z)
                        elif isinstance(picosCon, ConvexQuadraticConstraint):
                            picosDual = None
                        else:
                            assert False, "Unexpected constraint type."

                        # Flip sign based on objective sense.
                        if picosDual and self.ext.no.direction == "min":
                            picosDual = -picosDual
                    except cplex.exceptions.errors.CplexSolverError:
                        duals[picosCon] = None
                    else:
                        duals[picosCon] = picosDual

            # # Retrieve objective value.
            # try:
            #     if maxNumSolutions > 1:
            #         value = self.int.solution.pool.get_objective_value(
            #             solutionNum)
            #     else:
            #         value = self.int.solution.get_objective_value()
            # except cplex.exceptions.errors.CplexSolverError:
            #     value = None

            # Retrieve solution status.
            code = self.int.solution.get_status()
            if code in CPLEX_STATUS_CODES:
                prmlStatus, dualStatus, probStatus = CPLEX_STATUS_CODES[code]
            else:
                prmlStatus = SS_UNKNOWN
                dualStatus = SS_UNKNOWN
                probStatus = PS_UNKNOWN

            info = {}
            if o.cplex_bnd_monitor:
                info["bounds_monitor"] = callback.bounds

            solutions.append(self._make_solution(
                primals, duals, prmlStatus, dualStatus, probStatus, info))

        if maxNumSolutions > 1:
            return solutions
        else:
            assert len(solutions) == 1
            return solutions[0]


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Optimization problem modeling toolbox."""

from ..apidoc import api_end, api_start

_API_START = api_start(globals())
# -------------------------------


from .footprint  import Footprint, Specification  # noqa
from .options    import Options  # noqa
from .problem    import Objective, Problem, SolutionFailure  # noqa
from .solution   import Solution  # noqa
from .strategy   import NoStrategyFound, Strategy  # noqa
from .quicksolve import find_assignment, maximize, minimize  # noqa


# --------------------------------------
__all__ = api_end(_API_START, globals())
